# -*- coding: utf-8 -*-
# Example tutorial for writing a tag source plugin using Amazon's webservice API.

# The important stuff is at the bottom in the Amazon class.
# Comments (or you've created a tag source you'd like
# to share) can be directed at concentricpuddle@gmail.com

import base64
import hashlib
import hmac
import re
import time
import urllib.error
import urllib.parse
import urllib.request

from xml.dom import minidom

from puddlestuff.audioinfo import DATA
from puddlestuff.constants import CHECKBOX, COMBO, TEXT
from puddlestuff.tagsources import (write_log, RetrievalError,
                          urlopen, parse_searchstring)
from puddlestuff.util import translate

default_access_key = base64.b64decode('QUtJQUozS0JZUlVZUU41UFZRR0E=')
default_secret_key = base64.b64decode('dmh6Q0ZaSEF6N0VvMmN5REt3STVnS1liU3ZFTCtSckx3c0tmanZEdA==')

access_key = default_access_key
secret_key = default_secret_key

SMALLIMAGE = '#smallimage'
MEDIUMIMAGE = '#mediumimage'
LARGEIMAGE = '#largeimage'

image_types = [SMALLIMAGE, MEDIUMIMAGE, LARGEIMAGE]

XMLKEYS = {
    'Artist': 'artist',
    'Label': 'label',
    'ReleaseDate': 'year',
    'Title': 'album',
    "Publisher": 'publisher'}

IMAGEKEYS = {'SmallImage': SMALLIMAGE,
             'MediumImage': MEDIUMIMAGE,
             'LargeImage': LARGEIMAGE}


def check_binding(node):
    """Checks whether a returned item as an Audio CD."""
    try:
        binding = node.getElementsByTagName('Binding')[0].firstChild.data
        return binding == 'Audio CD'
    except IndexError:
        return


def create_aws_url(aws_access_key_id, secret, query_dictionary):
    """Creates the query url that'll be used to query Amazon's service."""
    query_dictionary["AWSAccessKeyId"] = aws_access_key_id
    query_dictionary["Timestamp"] = time.strftime("%Y-%m-%dT%H:%M:%SZ",
                                                  time.gmtime())

    items = [(key, value) for key, value in
             query_dictionary.items()]
    query = urllib.parse.urlencode(sorted(items))

    try:
        hm = hmac.new(secret, "GET\nwebservices.amazon.com\n/onca/xml\n" \
                      +query, hashlib.sha256)
    except TypeError:
        raise RetrievalError(translate('Amazon',
                                       'Invalid Access or Secret Key'))
    signature = urllib.parse.quote(base64.b64encode(hm.digest()))

    query = "http://webservices.amazon.com/onca/xml?%s&Signature=%s" % (
        query, signature)
    return query


def check_matches(albums, artist=None, album_name=None):
    ret = []
    if artist and album_name:
        album_name = album_name.lower()
        artist = artist.lower()

        ret = [album for album in albums if
               album_name in album.get('album', album_name).lower() and
               artist in album.get('artist', artist).lower()]
    elif artist:
        artist = artist.lower()
        ret = [
            album for album in albums if
            artist in album.get('artist', artist).lower()
        ]
    elif album_name:
        album_name = album_name.lower()
        ret = [
            album for album in albums if
            album_name in album.get('album', album).lower()
        ]

    return ret if ret else albums


def get_asin(node):
    """Retrieves the ASIN of a node."""
    return node.getElementsByTagName('ASIN')[0].firstChild.data


def get_image_url(node):
    return node.getElementsByTagName('URL')[0].firstChild.data


def get_site_url(node):
    return node.getElementsByTagName('DetailPageURL')[0].firstChild.data


def get_text(node):
    """Returns the textual data in a node."""
    return node.firstChild.data


def keyword_search(keywords):
    write_log(translate('Amazon',
                        "Retrieving search results for keywords: {}").format(keywords))
    query_pairs = {
        "Operation": "ItemSearch",
        'SearchIndex': 'Music',
        "ResponseGroup": "ItemAttributes,Images",
        "Service": "AWSECommerceService",
        'ItemPage': '1',
        'Keywords': keywords,
        'AssociateTag': 'puddletag-20'}
    url = create_aws_url(access_key, secret_key, query_pairs)
    xml = urlopen(url)
    return parse_search_xml(xml)


def parse_album_xml(text, album=None):
    """Parses the retrieved xml for an album and get's the track listing."""
    doc = minidom.parseString(text)
    album_item = doc.getElementsByTagName('Item')[0]
    try:
        tracklist = album_item.getElementsByTagName('Tracks')[0]
    except IndexError:
        write_log(translate('Amazon',
                            'No tracks found in listing.'))
        write_log(text)
        return None
    tracks = []
    discs = [disc for disc in tracklist.childNodes if
             not disc.nodeType == disc.TEXT_NODE]
    if not (len(discs) > 1 and album):
        album = None
    for discnum, disc in enumerate(discs):
        for track_node in disc.childNodes:
            if track_node.nodeType == track_node.TEXT_NODE:
                continue
            title = get_text(track_node)
            tracknumber = track_node.attributes['Number'].value
            if album:
                tracks.append({'track': tracknumber, 'title': title,
                               'album': '%s (Disc %s)' % (album, discnum + 1)})
            else:
                tracks.append({'track': tracknumber, 'title': title})
    return tracks


def parse_search_xml(text):
    """Parses the xml retrieved after entering a search query. Returns a
    list of the albums found.
    """
    doc = minidom.parseString(text)
    items = doc.getElementsByTagName('Item')
    ret = []

    for item in items:
        info = {}
        for attrib in item.getElementsByTagName('ItemAttributes'):
            if not check_binding(attrib):
                continue
            for child in attrib.childNodes:
                if child.nodeType == child.TEXT_NODE:
                    continue
                if child.tagName in XMLKEYS:
                    text_node = child.firstChild
                    if text_node.nodeType == text_node.TEXT_NODE:
                        info[XMLKEYS[child.tagName]] = text_node.data
        if not info:
            continue
        for key in IMAGEKEYS:
            image_items = item.getElementsByTagName(key)
            if image_items:
                info[IMAGEKEYS[key]] = get_image_url(image_items[0])
        info['#extrainfo'] = (translate('Amazon', "{} at Amazon.com").format(
                              info.get('album', '')), get_site_url(item))
        info['#asin'] = get_asin(item)
        info['asin'] = info['#asin']
        ret.append(info)
    return ret


def retrieve_album(info, image=MEDIUMIMAGE):
    """Retrieves album from the information in info.
    image must be either one of image_types or None.
    If None, no image is retrieved."""
    if isinstance(info, str):
        asin = info
    else:
        asin = info['#asin']

    query_pairs = {
        "Operation": "ItemLookup",
        "Service": "AWSECommerceService",
        'ItemId': asin,
        'ResponseGroup': 'Tracks',
        'AssociateTag': 'puddletag-20'}
    url = create_aws_url(access_key, secret_key, query_pairs)

    if isinstance(info, str):
        write_log(translate('Amazon',
                            "Retrieving using ASIN: {}").format(asin))
    else:
        write_log(translate('Amazon',
                            "Retrieving XML: {} - {}"
                            ).format(info.get('artist', ''), info.get('album', '')))
    xml = urlopen(url)

    if isinstance(info, str):
        tracks = parse_album_xml(xml)
    else:
        tracks = parse_album_xml(xml, info['album'])

    if image in image_types:
        url = info[image]
        write_log(translate('Amazon', "Retrieving cover: {}").format(url))
        info.update({'__image': retrieve_cover(url)})
    return tracks


def retrieve_cover(url):
    data = urlopen(url)
    return [{DATA: data}]


def search(artist=None, album=None):
    if artist and album:
        keywords = '+'.join([artist, album])
    elif artist:
        keywords = artist
    else:
        keywords = album
    keywords = re.sub(r'(\s+)', '+', keywords)
    return keyword_search(keywords)

# A couple of things you should be aware of.
# If you're retrieving urls use puddlestuff.tagsources.urlopen instead of the
# generic urllib functions. When I implement some progress bars, proxies, etc
# your tag sources will benefit automatically. Pass it your url and
# it'll return your data.

# Raise puddlestuff.tagsources.RetrievalError if an error occurs. It
# accepts a message the only argument (may be html).

# Images must be a list of  dictonaries with the
# following keys (all from puddlestuff.audioinfo):
# DATA: Required. The images string data.
# MIMETYPE: either 'image/jpg' or 'image/png'
# DESCRIPTION: the image's description
# IMAGETYPE: Index of the element corresponding to audioinfo.IMAGETYPES


# The only thing required is just a Tag Source object,
# which'll be the interface to puddletag.
class Amazon(object):
    # The name attribute is required.
    name = 'Amazon'
    # group_by specifies how the tag sources wants files to be grouped.
    # in this case they'll be grouped by album first and then by artists.

    # The values passed to Object.search will be album (a string)
    # and dictionary containing the different artists as keys and
    # a list of tags (dictionaries) as the values.
    group_by = ['album', 'artist']

    # So for these values the values passed to Object.search may be
    # ("Give Up", artist={'The Postal Service': [track listing...]})

    # Each element of track listing is a dictionary of field: value keys
    # containing file-specific info., eg.
    # {'track': ['10'], 'title': ['Natural Anthem'], '__length': '5:07'}
    # values are usually lists and "builtin" values are strings.
    # All strings are unicode, so don't worry about conversions.

    tooltip = translate('Amazon',
                        """<p>Enter search parameters here. If empty, the selected files
                        are used.</p>
                        <ul>
                        <li><b>artist;album</b>
                        searches for a specific album/artist combination.</li>
                        <li>To list the albums by an artist leave off the album part,
                        but keep the semicolon (eg. <b>Ratatat;</b>).
                        For a album only leave the artist part as in
                        <b>;Resurrection.</li>
                        <li>Entering keywords <b>without a semi-colon (;)</b> will do an
                        Amazon album search using those keywords.</li>
                        </ul>""")

    # __init__ should not accept any arguments.
    def __init__(self):
        super(Amazon, self).__init__()
        self._getcover = True
        self.covertype = 1

        # Object.preferences is a list of lists definining the the controls
        # that'll be used to created the Tag Source's config dialog.

        # Currently, there are three types of controls:
        # TEXT, COMBO and CHECKBOX. They correspond to QLineEdit,
        # QComboBox and QCheckBox respectively.

        # Each control requires three arguments in order to be created.
        # 1. Some descriptive text that's shown in a label.
        # 2. The control type, either TEXT, COMBO or CHECKBOX
        # 3. For TEXT this argument is the default text. It's not required.
        #   For COMBO, the default argument is a list containing a list
        #   of strings as the first item.
        #   And the default index as the second item. eg.
        #   [['text1', 'text2'], 1]
        #   Checkboxes can either be checked or not so
        #   default arguments must either True or False.

        # When the user has finished editing, the Object.applyPrefs will be
        # called with a list of values correcsponding to the order
        # in which the were defined.
        # The value returned will be either True or False for CHECKBOX.
        # For TEXT, it'll be just the text and for COMBO it'll be the index
        # the user's selected.

        # Values will be saved autotically and the applyPrefs method
        # will be called when puddletag starts.

        self.preferences = [
            [translate('Amazon', 'Retrieve Cover'), CHECKBOX, True],
            [translate('Amazon', 'Cover size to retrieve'), COMBO,
             [[translate('Amazon', 'Small'),
               translate('Amazon', 'Medium'),
               translate('Amazon', 'Large')], 1]],
            [translate('Amazon', 'Access Key (Stored '
                                 'as plain-text. Leave empty for default.)'), TEXT, ''],
            [translate('Amazon', 'Secret Key (Stored '
                                 'as plain-text. Leave empty for default.)'), TEXT, ''],
        ]

    def keyword_search(self, text):
        """Searches for albums by keywords, text."""
        # Should search for the keywords in text.
        # This method is optional, but recommended.

        # The format artist1;album1|artist2;album2 should be accepted.
        # Use the parse_searchstring method to separate text
        # into a list of artist album pairs as in:
        # [(artist1, album1), (artist2, album2)]
        try:
            params = parse_searchstring(text)
            artists = [params[0][0]]
            album = params[0][1]
        except RetrievalError:
            album = text
            artists = None
        return self.search(album, artists)

    def search(self, album, artists):
        # Required.
        # See group_by's explanation for an overview of the arguments
        # that'll be passed to this function.

        # It should return a list consisting of (albuminfo, tracklisting)
        # pairs.

        # albuminfo is a dictionary containing information applicable to
        # the whole album name like
        # {'artist': ['The Postal Service'], 'album': ['Give Up'],
        #    'year': '2003'}
        # Values can be either strings or lists of strings,
        # but all strings must be unicode!

        # Any key starting with '#' will be considered source specific and
        # will not be changed under any circumstances. Use them to store
        # any info that'll be used later to retrieve tracks.

        # Tracks use the same format as albums, but contain info specific to a
        # a track eg. [
        #    {'title': 'The District Sleeps tonight', 'track': '1'},
        #    {'title': 'Such Great Heights', '2'}
        #    .....
        #    ]

        # Return An empty list (info, []) if the tag source
        # doesn't include tracks with a lookup. Tracks can later
        # be retrieved at the user's request.

        # Do the same even if an exact match was found, but an extra
        # lookup is required to retrieve tracks.

        if artists:
            if len(artists) > 1:
                artist = 'Various Artists'
            else:
                try:
                    artist = list(artists.keys())[0]
                except AttributeError:
                    artist = artists[0]
        else:
            artist = None

        retrieved_albums = search(artist, album)
        matches = check_matches(retrieved_albums, artist, album)
        return [(info, []) for info in matches]

    def retrieve(self, info):
        # Required. Will retrieve track listing using
        # info previously obtained from Object.search.

        # Should return albuminfo (many tag sources return more info on
        # the second lookup) and tracklisting.
        if self._getcover:
            return info, retrieve_album(info, self.covertype)
        else:
            return info, retrieve_album(info, None)

    def applyPrefs(self, args):
        self._getcover = args[0]
        self.covertype = image_types[args[1]]
        global access_key
        global secret_key
        if args[2]:
            access_key = args[2]
            secret_key = args[3]
        else:
            access_key = default_access_key
            secret_key = default_secret_key


# Required in order to let your tagsource be loaded.
tagsources = [Amazon]
info = Amazon

if __name__ == '__main__':
    x = Amazon()
    print(x.keyword_search('amy winehouse'))



import struct

from mutagen.asf import ASF, ASFByteArrayAttribute, ASFUnicodeAttribute

from . import tag_versions
from . import util
from .util import (usertags, PATH,
                   FILENAME, isempty, getdeco, setdeco, str_filesize, fn_hash, cover_info,
                   get_mime, unicode_list, CaselessDict, keys_deco,
                   del_deco, get_total, set_total, info_to_dict, parse_image)

ATTRIBUTES = ('frequency', 'length', 'bitrate', 'accessed', 'size', 'created',
              'modified', 'filetype')


# From picard
def bin_to_pic(image):
    data = image.value
    (type, size) = struct.unpack_from("<bi", data)
    pos = 5
    mime = b""
    while data[pos:pos + 2] != b"\x00\x00":
        mime += data[pos:pos + 2]
        pos += 2
    pos += 2
    description = b""
    while data[pos:pos + 2] != b"\x00\x00":
        description += data[pos:pos + 2]
        pos += 2
    pos += 2
    image_data = data[pos:pos + size]
    return {
        util.MIMETYPE: mime.decode("utf-16-le"),
        util.DATA: image_data,
        util.IMAGETYPE: type,
        util.DESCRIPTION: description.decode("utf-16-le"),
    }


def pic_to_bin(image):
    data = image[util.DATA]
    mime = image.get(util.MIMETYPE)
    if not mime:
        mime = get_mime(data)
        if not mime:
            return
    type = image.get(util.IMAGETYPE, 3)
    description = image.get(util.DESCRIPTION, '')
    tag_data = struct.pack("<bi", type, len(data))
    tag_data += mime.encode("utf-16-le") + b"\x00\x00"
    tag_data += description.encode("utf-16-le") + b"\x00\x00"
    tag_data += data
    return ASFByteArrayAttribute(tag_data)


class Tag(util.MockTag):
    IMAGETAGS = (util.MIMETYPE, util.DATA, util.IMAGETYPE, util.DESCRIPTION)
    mapping = {}
    revmapping = {}

    __rtranslate = {
        'album': 'WM/AlbumTitle',
        'album_subtitle': 'WM/SetSubTitle',
        'albumartist': 'WM/AlbumArtist',
        'albumartistsortorder': 'WM/AlbumArtistSortOrder',
        'albumsortorder': 'WM/AlbumSortOrder',
        'artist': 'Author',
        'bpm': 'WM/BeatsPerMinute',
        'comment:': 'Description',
        'composer': 'WM/Composer',
        'conductor': 'WM/Conductor',
        'copyright': 'WM/Copyright',
        'discnumber': 'WM/PartOfSet',
        'encodedby': 'WM/EncodedBy',
        'encodersettings': 'WM/EncodingSettings',
        'encodingtime': 'WM/EncodingTime',
        'engineer': 'WM/Producer',
        'genre': 'WM/Genre',
        'grouping': 'WM/ContentGroupDescription',
        'initialkey': 'WM/InitialKey',
        'isrc': 'WM/ISRC',
        'label': 'WM/Publisher',
        'language': 'WM/Language',
        'lyricist': 'WM/Writer',
        'lyrics': 'WM/Lyrics',
        'mbrainz_album_id': 'MusicBrainz/Album Id',
        'mbrainz_albumartist_id': 'MusicBrainz/Album Artist Id',
        'mbrainz_artist_id': 'MusicBrainz/Artist Id',
        'mbrainz_discid': 'MusicBrainz/Disc Id',
        'mbrainz_track_id': 'MusicBrainz/Track Id',
        'mbrainz_trmid': 'MusicBrainz/TRM Id',
        'mood': 'WM/Mood',
        'musicip_puid': 'MusicIP/PUID',
        'originalalbum': 'WM/OriginalAlbumTitle',
        'originalartist': 'WM/OriginalArtist',
        'originalfilename': 'WM/OriginalFilename',
        'originallyricist': 'WM/OriginalLyricist',
        'originalyear': 'WM/OriginalReleaseYear',
        'performersortorder': 'WM/ArtistSortOrder',
        'releasecountry': 'MusicBrainz/Album Release Country',
        'releasestatus': 'MusicBrainz/Album Status',
        'releasetype': 'MusicBrainz/Album Type',
        'remixer': 'WM/ModifiedBy',
        'subtitle': 'WM/SubTitle',
        'title': 'Title',
        'titlesortorder': 'WM/TitleSortOrder',
        'track': 'WM/TrackNumber',
        'unsyncedlyrics': 'WM/Lyrics',
        'wwwartist': 'WM/AuthorURL',
        'wwwaudiosource': 'WM/AudioSourceURL',
        'wwwcommercialinfo': 'WM/PromotionURL',
        'wwwcopyright': 'CopyrightURL',
        'year': 'WM/Year',
    }
    __translate = dict([(v, k) for k, v in __rtranslate.items()])

    def __init__(self, filename=None):
        self.__images = []
        self.__tags = CaselessDict()

        util.MockTag.__init__(self, filename)

    @property
    def filepath(self):
        return util.MockTag.get_filepath(self)

    @filepath.setter
    def filepath(self, val):
        self.__tags.update(util.MockTag.set_filepath(self, val))

    @property
    def images(self):
        return self.__images

    @images.setter
    def images(self, images):
        if images:
            self.__images = list(map(parse_image, images))
        else:
            self.__images = []
        cover_info(images, self.__tags)

    def __contains__(self, key):
        if key == '__image':
            return bool(self.images)

        elif key == '__total':
            try:
                return bool(get_total(self))
            except (KeyError, ValueError):
                return False

        if self.revmapping:
            key = self.revmapping.get(key, key)
        return key in self.__tags

    @getdeco
    def __getitem__(self, key):
        if key == '__image':
            return self.images
        elif key == '__total':
            return get_total(self)
        return self.__tags[key]

    @setdeco
    def __setitem__(self, key, value):
        if key.startswith('__'):

            if key == '__image':
                self.images = value
            elif key in fn_hash:
                setattr(self, fn_hash[key], value)
            elif key == '__total' and 'track' in self:
                if set_total(self, value):
                    return
            return
        elif isempty(value):
            del (self[key])
        elif key in self.__rtranslate:
            self.__tags[key] = unicode_list(value)

    @del_deco
    def __delitem__(self, key):
        if key == '__image':
            self.images = []
        elif key.startswith('__'):
            return
        else:
            del (self.__tags[key])

    def delete(self):
        self.mut_obj.clear()
        for key in self.usertags:
            del (self.__tags[self.revmapping.get(key, key)])
        self.images = []
        self.save()

    @property
    def info(self):
        info = self.mut_obj.info
        fileinfo = [('Path', self[PATH]),
                    ('Size', str_filesize(int(self.size))),
                    ('Filename', self[FILENAME]),
                    ('Modified', self.modified)]

        wmainfo = [('Bitrate', self.bitrate),
                   ('Frequency', self.frequency),
                   ('Channels', str(info.channels)),
                   ('Length', self.length)]
        return [('File', fileinfo), ('ASF Info', wmainfo)]

    @keys_deco
    def keys(self):
        return list(self.__tags.keys())

    def link(self, filename):
        """Links the audio, filename
        returns self if successful, None otherwise."""

        self.images = None
        tags, audio = self.load(filename, ASF)
        if audio is None:
            return

        for name, values in audio.tags.items():
            try:
                self.__tags[self.__translate[name]] = list(map(str, values))
            except KeyError:
                if isinstance(values[0], ASFUnicodeAttribute):
                    if not '/' in name and name not in self.__tags:
                        self.__tags[name] = list(map(str, values))

        if 'WM/Picture' in audio:
            self.images = list(map(bin_to_pic, audio['WM/Picture']))

        self.__tags.update(info_to_dict(audio.info))
        self.__tags.update(tags)
        self.__tags['__filetype'] = 'ASF'
        self.filetype = 'ASF'
        self.__tags['__tag_read'] = 'ASF'
        self.set_attrs(ATTRIBUTES)
        self.mut_obj = audio
        self.update_tag_list()
        return self

    def save(self):
        """Writes the tags in self.__tags
        to self.filename if no filename is specified."""
        filepath = self.filepath

        if self.mut_obj.tags is None:
            self.mut_obj.add_tags()
        if filepath != self.mut_obj.filename:
            self.mut_obj.filename = filepath
        audio = self.mut_obj

        newtag = {}
        for field, value in usertags(self.__tags).items():
            try:
                newtag[self.__rtranslate[field]] = value
            except KeyError:
                newtag[field] = value

        pics = list(map(pic_to_bin, self.images))
        if pics:
            newtag['WM/Picture'] = pics

        toremove = [z for z in audio if z not in newtag]
        for z in toremove:
            del (audio[z])
        audio.update(newtag)
        audio.save()

    def update_tag_list(self):
        l = tag_versions.tags_in_file(self.filepath)
        if l:
            self.__tags['__tag'] = 'ASF, ' + ', '.join(l)
        else:
            self.__tags['__tag'] = 'ASF'


filetype = (ASF, Tag, 'WMA', ['wma', 'wmv'])



import os
import sys

from PyQt5.QtCore import QEvent, QLineF, QRectF, Qt, pyqtRemoveInputHook
from PyQt5.QtGui import QBrush, QKeySequence, QPainter, QPalette, QPen
from PyQt5.QtWidgets import qApp, QAbstractItemDelegate, QAbstractItemView, QApplication, QFrame, QItemDelegate, QLabel, \
    QTableWidget, QTableWidgetItem, QVBoxLayout, QWidget

from . import loadshortcuts as ls
from .constants import CONFIGDIR
from .puddleobjects import PuddleConfig
from .translations import translate

pyqtRemoveInputHook()


class ActionEditorWidget(QLabel):

    # Redefine the tr() function for this class.
    def tr(self, text):

        return qApp.translate("ActionEditorWidget", text)

    def __init__(self, text, parent):

        QLabel.__init__(self, text, parent)
        self.key = ""
        self.modifiers = {}
        self.setAutoFillBackground(True)
        palette = self.palette()
        palette.setBrush(QPalette.ColorRole.Base, palette.brush(QPalette.ColorRole.AlternateBase))
        self.setPalette(palette)
        self.valid = False
        self.setFrameStyle(QFrame.Shape.Panel)

    def keyPressEvent(self, event):

        other = None

        if event.key() == Qt.Key.Key_Shift:
            self.modifiers[Qt.Key.Key_Shift] = "Shift"
        elif event.key() == Qt.Key.Key_Control:
            self.modifiers[Qt.Key.Key_Control] = "Ctrl"
        elif event.key() == Qt.Key.Key_Meta:
            self.modifiers[Qt.Key.Key_Meta] = "Meta"
        elif event.key() == Qt.Key.Key_Alt:
            self.modifiers[Qt.Key.Key_Alt] = "Alt"
        else:
            other = QKeySequence(event.key()).toString(QKeySequence.SequenceFormat.NativeText)

        if other:
            key_string = "+".join(list(self.modifiers.values()) + [str(other), ])
            self.valid = True
        else:
            key_string = "+".join(list(self.modifiers.values()))

        self.setText(key_string)

    def keyReleaseEvent(self, event):

        if self.valid:
            return

        if event.key() == Qt.Key.Key_Shift:
            if Qt.Key.Key_Shift in self.modifiers:
                del self.modifiers[Qt.Key.Key_Shift]
        elif event.key() == Qt.Key.Key_Control:
            if Qt.Key.Key_Control in self.modifiers:
                del self.modifiers[Qt.Key.Key_Control]
        elif event.key() == Qt.Key.Key_Meta:
            if Qt.Key.Key_Meta in self.modifiers:
                del self.modifiers[Qt.Key.Key_Meta]
        elif event.key() == Qt.Key.Key_Alt:
            if Qt.Key.Key_Alt in self.modifiers:
                del self.modifiers[Qt.Key.Key_Alt]

        self.setText("+".join(list(self.modifiers.values())))

        if len(self.modifiers) == 0:
            self.releaseKeyboard()

    def mousePressEvent(self, event):

        if event.button() != Qt.MouseButton.LeftButton:
            return

        size = self.height() / 2.0
        rect = QRectF(self.width() - size, size * 0.5, size, size)

        if rect.contains(event.localPos()):
            self.clear()
            self.valid = True
            event.accept()

    def paintEvent(self, event):

        if self.text():
            painter = QPainter()
            painter.begin(self)
            painter.setRenderHint(QPainter.RenderHint.Antialiasing)

            color = self.palette().color(QPalette.ColorRole.Highlight)
            color.setAlpha(127)
            painter.setBrush(QBrush(color))
            color = self.palette().color(QPalette.ColorRole.HighlightedText)
            color.setAlpha(127)
            painter.setPen(QPen(color))
            size = self.height() / 2.0

            left = self.width() - 4

            painter.drawRect(QRectF(left - size, size * 0.5, size, size))
            painter.drawLine(QLineF(left - size * 0.75, size * 0.75,
                             left - size * 0.25, size * 1.25))
            painter.drawLine(QLineF(left - size * 0.25, size * 0.75,
                             left - size * 0.75, size * 1.25))
            painter.end()

        QLabel.paintEvent(self, event)

    def showEvent(self, event):

        self.grabKeyboard()


class ActionEditorDelegate(QItemDelegate):

    def __init__(self, parent=None):

        QItemDelegate.__init__(self, parent)

    def createEditor(self, parent, option, index):

        self._edited = str(index.data())

        self.editor = ActionEditorWidget(str(index.data()), parent)
        self.editor.installEventFilter(self)
        return self.editor

    def eventFilter(self, obj, event):

        if obj == self.editor:
            if event.type() == QEvent.Type.KeyPress:
                obj.keyPressEvent(event)
                if obj.valid:
                    self.commitData.emit(self.editor)
                    self.closeEditor.emit(self.editor, QAbstractItemDelegate.EndEditHint.NoHint)
                return True

            elif event.type() == QEvent.Type.KeyRelease:
                obj.keyReleaseEvent(event)
                if not obj.text():
                    self.closeEditor.emit(self.editor, QAbstractItemDelegate.EndEditHint.NoHint)
                return True

            elif event.type() == QEvent.Type.MouseButtonPress:
                obj.mousePressEvent(event)
                if obj.valid:
                    self.commitData.emit(self.editor)
                    self.closeEditor.emit(self.editor, QAbstractItemDelegate.EndEditHint.NoHint)
                return True

        return False

    def paint(self, painter, option, index):

        if index.column() != 0:
            QItemDelegate.paint(self, painter, option, index)
            return

        painter.fillRect(option.rect, option.palette.brush(QPalette.ColorRole.Base))
        painter.setPen(QPen(option.palette.color(QPalette.ColorRole.Text)))
        painter.drawText(option.rect.adjusted(4, 4, -4, -4),
                         Qt.TextFlag.TextShowMnemonic | Qt.AlignmentFlag.AlignLeft | Qt.AlignmentFlag.AlignVCenter,
                         str(index.data()))

    def setEditorData(self, editor, index):

        editor.setText(str(index.data()))

    def setModelData(self, editor, model, index):
        if editor.text() != self._edited:
            index.model().edited = True
        model.setData(index, editor.text())

    def updateEditorGeometry(self, editor, option, index):
        editor.setGeometry(option.rect)


class ActionEditorDialog(QWidget):

    # Redefine the tr() function for this class.
    def tr(self, text):

        return qApp.translate("ActionEditorDialog", text)

    def __init__(self, actions, parent=None):

        super(ActionEditorDialog, self).__init__(parent)
        self.actions = actions

        help = QLabel(translate("Shortcut Settings", '<b>Double click a cell in the Shortcut Column'
                                                     ' to <br />modify the key sequence.</b>'))

        self.actionTable = QTableWidget(self)
        self.actionTable.setSelectionBehavior(QAbstractItemView.SelectionBehavior.SelectRows)
        self.actionTable.setEditTriggers(QAbstractItemView.EditTrigger.DoubleClicked)
        self.actionTable.setColumnCount(2)
        self.actionTable.setHorizontalHeaderLabels(
            [translate("Shortcut Settings", "Description"),
             translate("Shortcut Settings", "Shortcut")]
        )
        self.actionTable.horizontalHeader().setStretchLastSection(True)
        self.actionTable.verticalHeader().hide()
        self.actionTable.setItemDelegate(ActionEditorDelegate(self))

        self.actionTable.cellChanged.connect(self.validateAction)

        row = 0
        for action in self.actions:

            if not action.text():
                continue

            self.actionTable.insertRow(self.actionTable.rowCount())

            item = QTableWidgetItem()
            item.setText(action.text())
            item.setFlags(Qt.ItemFlag.ItemIsEnabled | Qt.ItemFlag.ItemIsSelectable)
            self.actionTable.setItem(row, 0, item)

            item = QTableWidgetItem()
            item.setText(action.shortcut().toString())
            item.setFlags(Qt.ItemFlag.ItemIsEnabled | Qt.ItemFlag.ItemIsEditable | Qt.ItemFlag.ItemIsSelectable)
            item.oldShortcutText = item.text()
            self.actionTable.setItem(row, 1, item)

            row += 1

        self.actionTable.resizeColumnsToContents()

        mainLayout = QVBoxLayout()
        mainLayout.addWidget(help)
        mainLayout.setContentsMargins(8, 8, 8, 8)
        mainLayout.setSpacing(8)
        mainLayout.addWidget(self.actionTable)
        self.setLayout(mainLayout)
        self._model = self.actionTable.model()
        self._model.edited = False
        self.actionTable.model().edited = False

        self.setWindowTitle(translate("Shortcut Settings", "Edit Shortcuts"))

    def applySettings(self, control=None):
        if not self._model.edited:
            return

        row = 0
        for action in self.actions:

            if action.text():
                action.setText(self.actionTable.item(row, 0).text())
                action.setShortcut(QKeySequence(self.actionTable.item(row, 1).text()))
                row += 1
        self.saveSettings(self.actions)
        self._model.edited = False

    def _loadSettings(self, actions):

        cparser = PuddleConfig(os.path.join(CONFIGDIR, 'user_shortcuts'))

        for action in actions:
            shortcut = cparser.get('shortcuts', str(action.text()), '')
            if shortcut:
                action.setShortcut(QKeySequence(shortcut))

    _loadSettings = classmethod(_loadSettings)

    def saveSettings(self, actions):

        cparser = PuddleConfig(os.path.join(CONFIGDIR, 'user_shortcuts'))
        for action in actions:
            shortcut = str(action.shortcut().toString())
            cparser.set('shortcuts', str(action.text()), shortcut)

    saveSettings = classmethod(saveSettings)

    def validateAction(self, row, column):

        if column != 1:
            return

        item = self.actionTable.item(row, column)
        shortcutText = QKeySequence(item.text()).toString()
        thisRow = self.actionTable.row(item)

        if shortcutText:
            for row in range(self.actionTable.rowCount()):
                if row == thisRow:
                    continue

                other = self.actionTable.item(row, 1)

                if other.text() == shortcutText:
                    other.setText(item.oldShortcutText)
                    break

            item.setText(shortcutText)
            item.oldShortcutText = shortcutText

        self.actionTable.resizeColumnToContents(1)


if __name__ == '__main__':
    app = QApplication(sys.argv)
    # win = ShortcutSettings()
    # win = EditShortcut('Open', 'Ctrl+O')
    widget = QWidget()
    actions = ls.get_actions(widget)
    ActionEditorDialog._loadSettings(actions)
    win = ActionEditorDialog(actions)
    # win.loadSettings(actions)
    win.show()
    app.exec_()



import glob
import os
from copy import deepcopy
from functools import partial

from PyQt5.QtCore import QMutex, QObject, pyqtSignal
from PyQt5.QtWidgets import QApplication, QCheckBox, QComboBox, QDialog, QGridLayout, QHBoxLayout, QLabel, \
    QLineEdit, QPushButton, QSpinBox, QTextEdit, QVBoxLayout, QWidget

from .. import masstag as masstag_module
from ..constants import RIGHTDOCK
from ..masstag import (NO_MATCH_OPTIONS, fields_from_text, match_files, masstag, merge_tsp_tracks,
                       split_files, MassTagFlag, MassTagProfile, TagSourceProfile)
from ..masstag.config import (PROFILEDIR, convert_mtps,
                              load_all_mtps, save_mtp)
from ..puddleobjects import (create_buddy, winsettings, ListButtons, ListBox, OKCancel, PuddleConfig, PuddleThread)
from ..translations import translate


class _SignalObject(QObject):
    statusChanged = pyqtSignal(str, name='statusChanged')
    logappend = pyqtSignal(str, name='logappend')


status_obj = _SignalObject()


def set_status(msg):
    status_obj.statusChanged.emit(msg)
    QApplication.processEvents()


# monkey-patch the module to use the gui output
masstag_module.set_status = set_status

mutex = QMutex()


def search_error(error, profile):
    set_status(translate('Masstagging',
                         "An error occured during the search: <b>{}</b>").format(str(error)))


def retrieval_error(error, profile):
    set_status(translate('Masstagging',
                         "An error occured during album retrieval: <b>{}</b>").format(str(error)))


class MassTagEdit(QDialog):
    profilesChanged = pyqtSignal(list, name='profilesChanged')

    def __init__(self, tag_sources, profiles=None, parent=None):
        super(MassTagEdit, self).__init__(parent)

        self.setWindowTitle(translate('Profile Config',
                                      'Configure Mass Tagging Profiles'))
        winsettings('masstag_edit', self)

        self.listbox = ListBox()
        self.tag_sources = tag_sources

        okcancel = OKCancel()
        okcancel.okButton.setDefault(True)

        self.buttonlist = ListButtons()

        connect = lambda control, signal, slot: getattr(control, signal).connect(slot)

        connect(okcancel, "ok", self.okClicked)
        connect(okcancel, "cancel", self.close)
        connect(self.buttonlist, "add", self.addClicked)
        connect(self.buttonlist, "edit", self.editClicked)
        connect(self.buttonlist, "moveup", self.moveUp)
        connect(self.buttonlist, "movedown", self.moveDown)
        connect(self.buttonlist, "remove", self.remove)
        connect(self.buttonlist, "duplicate", self.dupClicked)
        connect(self.listbox, "itemDoubleClicked", self.editClicked)
        connect(self.listbox, "currentRowChanged", self.enableListButtons)

        self.enableListButtons(self.listbox.currentRow())

        layout = QVBoxLayout()
        self.setLayout(layout)

        layout.addWidget(QLabel(translate('Profile Config',
                                          'Masstagging Profiles')))

        list_layout = QHBoxLayout()
        list_layout.addWidget(self.listbox, 1)
        list_layout.addLayout(self.buttonlist)

        layout.addLayout(list_layout)
        layout.addLayout(okcancel)
        if profiles is not None:
            self.setProfiles(profiles)

    def addClicked(self):
        win = MTProfileEdit(self.tag_sources, parent=self)
        win.setModal(True)
        win.profileChanged.connect(self.addProfile)
        win.show()

    def addProfile(self, profile):
        row = self.listbox.count()
        self.listbox.addItem(profile.name)
        self._profiles.append(profile)

    def dupClicked(self):
        row = self.listbox.currentRow()
        if row == -1:
            return
        win = MTProfileEdit(self.tag_sources, deepcopy(self._profiles[row]), self)
        win.setModal(True)
        win.profileChanged.connect(self.addProfile)
        win.show()

    def editClicked(self, item=None):
        if item:
            row = self.listbox.row(item)
        else:
            row = self.listbox.currentRow()

        if row == -1:
            return
        win = MTProfileEdit(self.tag_sources, self._profiles[row], self)
        win.setModal(True)
        win.profileChanged.connect(partial(self.replaceProfile, row))
        win.show()

    def enableListButtons(self, val):
        if val == -1:
            [button.setEnabled(False) for button in self.buttonlist.widgets[1:]]
        else:
            [button.setEnabled(True) for button in self.buttonlist.widgets[1:]]

    def loadProfiles(self, dirpath=PROFILEDIR):
        profiles = load_all_mtps(dirpath, self.tag_sources)
        self.setProfiles([_f for _f in profiles if _f])

    def moveDown(self):
        self.listbox.moveDown(self._profiles)

    def moveUp(self):
        self.listbox.moveUp(self._profiles)

    def okClicked(self):
        self.profilesChanged.emit(self._profiles)
        self.saveProfiles(PROFILEDIR, self._profiles)
        self.close()

    def remove(self):
        row = self.listbox.currentRow()
        if row == -1:
            return
        del (self._profiles[row])
        self.listbox.takeItem(row)

    def replaceProfile(self, row, profile):
        self._profiles[row] = profile
        self.listbox.item(row).setText(profile.name)

    def saveProfiles(self, dirpath, profiles):
        filenames = {}
        order = []
        for profile in profiles:
            filename = profile.name + '.mtp'
            i = 0
            while filename in filenames:
                filename = '%s_%d%s' % (profile.name, i, '.mtp')
                i += 1
            filenames[filename] = profile
            order.append(profile.name)
        files = glob.glob(os.path.join(dirpath, '*.mtp'))
        for f in files:
            if f not in filenames:
                try:
                    os.remove(f)
                except EnvironmentError:
                    pass
        for filename, profile in filenames.items():
            save_mtp(profile, os.path.join(dirpath, filename))
        f = open(os.path.join(dirpath, 'order'), 'w')
        f.write('\n'.join(order))
        f.close()

    def setProfiles(self, profiles):
        self._profiles = profiles
        for profile in self._profiles:
            self.listbox.addItem(profile.name)


class MTProfileEdit(QDialog):
    profileChanged = pyqtSignal(MassTagProfile, name='profileChanged')

    def __init__(self, tag_sources, profile=None, parent=None):
        super(MTProfileEdit, self).__init__(parent)

        self.setWindowTitle(translate('Profile Editor', 'Edit Masstagging Profile'))
        winsettings('masstag_profile', self)
        self._configs = []
        self.tag_sources = tag_sources
        self._tsps = []

        self._name = QLineEdit(translate('Profile Editor',
                                         'Masstagging Profile'))

        self._desc = QLineEdit()

        self.listbox = ListBox()

        self.okcancel = OKCancel()
        self.okcancel.okButton.setDefault(True)

        self.buttonlist = ListButtons()

        self.pattern = QLineEdit('%artist% - %album%/%track% - %title%')
        self.pattern.setToolTip(translate('Profile Editor',
                                          "<p>If no tag information is found in a file, "
                                          "the tags retrieved using this pattern will be used instead.</p>"))

        self.albumBound = QSpinBox()
        self.albumBound.setRange(0, 100)
        self.albumBound.setValue(70)
        self.albumBound.setToolTip(translate('Profile Editor',
                                             "<p>The artist and album fields will be used in "
                                             "determining whether an album matches the retrieved one. "
                                             "Each field will be compared using a fuzzy matching algorithm. "
                                             "If the resulting average match percentage is greater "
                                             "or equal than what you specify here it'll be "
                                             "considered to match.</p>"))

        self.matchFields = QLineEdit('artist, title')
        self.matchFields.setToolTip(translate('Profile Editor',
                                              '<p>The fields listed here will be used in '
                                              'determining whether a file matches a retrieved track. '
                                              'Each field will be compared using a fuzzy matching '
                                              'algorithm. If the resulting average match '
                                              'percentage is greater than the "Minimum Percentage" '
                                              'it\'ll be considered to match.</p>'))

        self.trackBound = QSpinBox()
        self.trackBound.setRange(0, 100)
        self.trackBound.setValue(80)

        self.jfdi = QCheckBox(translate('Profile Editor',
                                        'Brute force unmatched files.'))
        self.jfdi.setToolTip(translate('Profile Editor',
                                       "<p>Check to enable brute forcing matches. "
                                       " If a proper match isn't found for a file, "
                                       'the files will get sorted by filename, '
                                       'the retrieved tag sources by filename and '
                                       'corresponding (unmatched) tracks will matched.</p>'))

        self.existing = QCheckBox(translate('Profile Editor',
                                            'Update empty fields only.'))

        self.grid = QGridLayout()
        self.setLayout(self.grid)

        self.grid.addLayout(create_buddy(
            translate('Profile Editor', '&Name:'), self._name), 0, 0, 1, 2)
        self.grid.addLayout(create_buddy(
            translate('Profile Editor', '&Description'), self._desc), 1, 0, 1, 2)
        self.grid.addWidget(self.listbox, 2, 0)
        self.grid.setRowStretch(2, 1)
        self.grid.addLayout(self.buttonlist, 2, 1)
        self.grid.addLayout(create_buddy(translate('Profile Editor',
                                                   'Pattern to match filenames against.'),
                                         self.pattern, QVBoxLayout()), 3, 0, 1, 2)
        self.grid.addLayout(create_buddy(translate('Profile Editor',
                                                   'Minimum percentage required for album matches.'),
                                         self.albumBound), 4, 0, 1, 2)
        self.grid.addLayout(create_buddy(translate('Profile Editor',
                                                   'Match tracks using fields: '),
                                         self.matchFields, QVBoxLayout()), 5, 0, 1, 2)
        self.grid.addLayout(create_buddy(translate('Profile Editor',
                                                   'Minimum percentage required for track match.'),
                                         self.trackBound), 6, 0, 1, 2)
        self.grid.addWidget(self.jfdi, 7, 0, 1, 2)
        self.grid.addWidget(self.existing, 8, 0, 1, 2)
        self.grid.addLayout(self.okcancel, 9, 0, 1, 2)

        connect = lambda control, signal, slot: getattr(control, signal).connect(slot)

        connect(self.okcancel, "ok", self.okClicked)
        connect(self.okcancel, "cancel", self.close)
        connect(self.buttonlist, "add", self.addClicked)
        connect(self.buttonlist, "edit", self.editClicked)
        connect(self.buttonlist, "moveup", self.moveUp)
        connect(self.buttonlist, "movedown", self.moveDown)
        connect(self.buttonlist, "remove", self.remove)
        connect(self.buttonlist, "duplicate", self.dupClicked)
        connect(self.listbox, "itemDoubleClicked", self.editClicked)
        connect(self.listbox, "currentRowChanged", self.enableListButtons)

        if profile is not None:
            self.setProfile(profile)
        self.enableListButtons(self.listbox.currentRow())

    def addClicked(self):
        win = TSProfileEdit(self.tag_sources, None, self)
        win.setModal(True)
        win.profileChanged.connect(self.addTSProfile)
        win.show()

    def addTSProfile(self, profile):
        row = self.listbox.count()
        self.listbox.addItem(profile.tag_source.name)
        self._tsps.append(profile)

    def dupClicked(self):
        row = self.listbox.currentRow()
        if row == -1:
            return
        win = TSProfileEdit(self.tag_sources, self._tsps[row], self)
        win.setModal(True)
        win.profileChanged.connect(self.addTSProfile)
        win.show()

    def editClicked(self, item=None):
        if item:
            row = self.listbox.row(item)
        else:
            row = self.listbox.currentRow()

        if row == -1:
            return
        win = TSProfileEdit(self.tag_sources, self._tsps[row], self)
        win.setModal(True)
        win.profileChanged.connect(partial(self.replaceTSProfile, row))
        win.show()

    def replaceTSProfile(self, row, profile):
        self._tsps[row] = profile
        self.listbox.item(row).setText(profile.tag_source.name)

    def enableListButtons(self, val):
        if val == -1:
            [button.setEnabled(False) for button in self.buttonlist.widgets[1:]]
        else:
            [button.setEnabled(True) for button in self.buttonlist.widgets[1:]]

    def moveDown(self):
        self.listbox.moveDown(self._tsps)

    def moveUp(self):
        self.listbox.moveUp(self._tsps)

    def okClicked(self):
        fields = [z.strip() for z in
                  str(self.matchFields.text()).split(',')]

        mtp = MassTagProfile(str(self._name.text()),
                             str(self._desc.text()), fields, None,
                             str(self.pattern.text()), self._tsps,
                             self.albumBound.value() / 100.0,
                             self.trackBound.value() / 100.0, self.jfdi.isChecked(),
                             self.existing.isChecked(), '')

        self.profileChanged.emit(mtp)
        self.close()

    def remove(self):
        row = self.listbox.currentRow()
        if row == -1:
            return
        del (self._tsps[row])
        self.listbox.takeItem(row)

    def setProfile(self, profile):
        self._tsps = [tsp for tsp in profile.profiles if tsp.tag_source]
        [self.listbox.addItem(tsp.tag_source.name) for tsp in self._tsps]
        self.albumBound.setValue(int(profile.album_bound * 100))
        self.pattern.setText(profile.file_pattern)
        self.matchFields.setText(', '.join(profile.fields))
        self.trackBound.setValue(int(profile.track_bound * 100))
        self.jfdi.setChecked(profile.jfdi)
        self._name.setText(profile.name)
        self._desc.setText(profile.desc)
        self.existing.setChecked(profile.leave_existing)


class TSProfileEdit(QDialog):
    profileChanged = pyqtSignal(TagSourceProfile, name='profileChanged')

    def __init__(self, tag_sources, profile=None, parent=None):
        super(TSProfileEdit, self).__init__(parent)
        self.setWindowTitle(translate('Profile Editor',
                                      'Edit Tag Source Config'))
        winsettings('ts_profile_edit', self)
        self.tag_sources = tag_sources

        layout = QVBoxLayout(self)
        self.setLayout(layout)

        self.fields = QLineEdit(self)
        layout.addLayout(
            create_buddy(translate('Defaults', 'Fields: '), self.fields))
        self.fields.setToolTip(translate('Profile Editor',
                                         '<p>Enter a comma-seperated list of fields to retrieve. '
                                         'Leave empty to retrieve all available fields/values. '

                                         '<br /><br />Eg. <b>artist, album, title</b> will '
                                         'only retrieve the artist, album and title fields of '
                                         'from the Tag Source. '

                                         '<br /><br />Start the list with '
                                         'the tilde (~) character to write all retrieved fields '
                                         ', but the ones listed. Eg the field list '
                                         '<b>~composer,__image</b> will write all fields but the '
                                         'composer and __image (aka cover art) fields.</p>'

                                         '<p>If a field has been retrieved in a previous '
                                         'Tag Source the values will be combined if they differ. '
                                         'Eg. If genre=<b>Rock</b> for the first tag source polled '
                                         'and genre=<b>Alternative</b> for the tag source '
                                         'polled second then the resulting field will have '
                                         'multiple-values ie. genre=<b>Rock\\\\Rap</b>'))

        self.replace_fields = QLineEdit(self)
        layout.addLayout(create_buddy(translate('Profile Editor',
                                                'Fields to replace: '), self.replace_fields))
        self.replace_fields.setToolTip(translate('Profile Editor',
                                                 'Enter a comma-separated lists of fields that\'ll replace any '
                                                 'retrieved from previously polled tag sources. '

                                                 '<br />Start the list with the tilde (~) character to replace all '
                                                 'but the fields you list. <br />'
                                                 '<b>NB: Fields listed here must also be listed in the '
                                                 'list of fields to retrieve.</b>'

                                                 '<br /><br />Eg. Assume you have two Tag Sources. '
                                                 'The first retrieves <b>artist=Freshlyground, '
                                                 'album=Nomvula, genre=Afro Pop</b>. The second source gets '
                                                 '<b>artist=Freshly Ground, album=Nomvula, '
                                                 'genre=Pop</b>. '
                                                 'For the second Tag Source, setting just <b>artist</b> as the '
                                                 'list of fields to replace will overwrite the artist '
                                                 'field retrieved from the first tag source. '
                                                 'The resulting retrieved fields/values as shown in puddletag will '
                                                 'then be <b>artist=Freshly Ground, album=Nomvula, '
                                                 'genre=Afro Pop\\\\Pop</b>.'))

        self.source = QComboBox()
        self.source.addItems([source.name for source in tag_sources])
        layout.addLayout(create_buddy(
            translate('Profile Editor', '&Source'), self.source))

        self.no_match = QComboBox()
        self.no_match.addItems(NO_MATCH_OPTIONS)
        layout.addLayout(create_buddy(translate('Profile Editor',
                                                '&If no results found: '), self.no_match))
        self.no_match.setToolTip(translate('Profile Editor',
                                           '<p><b>Continue</b>: The lookup for the current album will continue '
                                           'by checking the other tag sources if no matching results '
                                           'were found for this tag source.</p>'

                                           '<p><b>Stop:</b> The lookup for the current album will '
                                           'stop and any previously retrieved results will be used.</p>'))

        okcancel = OKCancel()
        okcancel.ok.connect(self._okClicked)
        okcancel.cancel.connect(self.close)
        layout.addLayout(okcancel)

        layout.addStretch()
        self.setMaximumHeight(layout.sizeHint().height())

        if profile:
            self.setProfile(profile)

    def _okClicked(self):
        source = self.tag_sources[self.source.currentIndex()]
        no_result = self.no_match.currentIndex()
        fields = fields_from_text(str(self.fields.text()))
        replace_fields = fields_from_text(str(self.replace_fields.text()))

        profile = TagSourceProfile(None, source, fields, no_result,
                                   replace_fields)

        self.close()
        self.profileChanged.emit(profile)

    def setProfile(self, profile):
        source_index = self.source.findText(profile.tag_source.name)
        if source_index != -1:
            self.source.setCurrentIndex(source_index)
        self.no_match.setCurrentIndex(profile.if_no_result)
        self.fields.setText(', '.join(profile.fields))
        self.replace_fields.setText(', '.join(profile.replace_fields))


class MassTagWindow(QWidget):
    setpreview = pyqtSignal(dict, name='setpreview')
    clearpreview = pyqtSignal(name='clearpreview')
    enable_preview_mode = pyqtSignal(name='enable_preview_mode')
    writepreview = pyqtSignal(name='writepreview')
    disable_preview_mode = pyqtSignal(name='disable_preview_mode')

    def __init__(self, parent=None, status=None):
        super(MassTagWindow, self).__init__(parent)
        self.receives = []
        self.emits = ['setpreview', 'clearpreview', 'enable_preview_mode',
                      'writepreview', 'disable_preview_mode']
        self.__flag = MassTagFlag()
        self.__flag.stop = False

        self.setWindowTitle(translate('Masstagging', 'Mass Tagging'))
        winsettings('masstaglog', self)
        self._startButton = QPushButton(translate('Masstagging', '&Search'))
        configure = QPushButton(translate('Masstagging', '&Configure Profiles'))
        write = QPushButton(translate('Masstagging', '&Write Previews'))
        clear = QPushButton(translate('Masstagging', 'Clear &Preview'))
        self._log = QTextEdit()
        self.tag_sources = status['initialized_tagsources']

        self.profileCombo = QComboBox()
        self.profile = None

        self._status = status

        status_obj.statusChanged.connect(self._appendLog)
        status_obj.logappend.connect(self._appendLog)
        self.profileCombo.currentIndexChanged.connect(self.changeProfile)
        self._startButton.clicked.connect(self.lookup)
        configure.clicked.connect(self.configure)
        write.clicked.connect(self.writePreview)
        clear.clicked.connect(self.clearPreview)

        buttons = QHBoxLayout()
        buttons.addWidget(configure)
        buttons.addStretch()
        buttons.addWidget(write)
        buttons.addWidget(clear)

        combo = QHBoxLayout()
        label = QLabel(translate('Masstagging', '&Profile:'))
        label.setBuddy(self.profileCombo)
        combo.addWidget(label)
        combo.addWidget(self.profileCombo, 1)
        combo.addWidget(self._startButton)

        layout = QVBoxLayout()
        layout.addLayout(combo)
        layout.addWidget(self._log)
        layout.addLayout(buttons)
        self.setLayout(layout)

    def _appendLog(self, text):
        mutex.lock()
        if not isinstance(text, str):
            text = str(text, 'utf8', 'replace')
        if text.startswith(':insert'):
            text = text[len(':insert'):]
            pos = len(self._log.toPlainText()) - 1
            pos = 0 if pos < 0 else pos
            self._log.textCursor().setPosition(pos)
            self._log.insertHtml(text)
        else:
            pos = len(self._log.toPlainText())
            if pos < 0:
                pos = 0
            self._log.textCursor().setPosition(pos)
            self._log.append(text)
        mutex.unlock()

    def changeProfile(self, index):
        try:
            self.profile = self.profiles[index]
        except IndexError:
            return
        cparser = PuddleConfig()
        cparser.set('masstagging', 'lastindex', index)

    def clearPreview(self):
        self.disable_preview_mode.emit()

    def configure(self):
        win = MassTagEdit(self.tag_sources, self.profiles, self)
        win.setModal(True)
        win.profilesChanged.connect(self.setProfiles)
        win.show()

    def lookup(self):
        button = self.sender()
        if self._startButton.text() != translate('Masstagging', '&Stop'):
            self.__flag.stop = False
            self._log.clear()
            self._startButton.setText(translate('Masstagging', '&Stop'))
            self._start()
        else:
            self._startButton.setText(translate('Masstagging', '&Search'))
            self.__flag.stop = True

    def loadSettings(self):
        convert_mtps(PROFILEDIR)
        if not os.path.exists(PROFILEDIR):
            os.mkdir(PROFILEDIR)
        self.setProfiles(load_all_mtps(PROFILEDIR, self.tag_sources))
        index = PuddleConfig().get('masstagging', 'lastindex', 0)
        if index < self.profileCombo.count():
            self.profileCombo.setCurrentIndex(index)

    def setProfiles(self, profiles):
        self.profiles = profiles
        if not profiles:
            self.profileCombo.clear()
            return
        self.profileCombo.currentIndexChanged.disconnect(self.changeProfile)
        old = self.profileCombo.currentText()
        self.profileCombo.clear()
        self.profileCombo.addItems([p.name for p in profiles])
        index = self.profileCombo.findText(old)
        if index == -1:
            index = 0
        self.profileCombo.setCurrentIndex(index)
        self.profile = profiles[index]

        if self.profile.desc:
            self.profileCombo.setToolTip(self.profile.desc)
        self.profileCombo.currentIndexChanged.connect(self.changeProfile)

    def _start(self):
        mtp = self.profile
        if self.profile == None:
            set_status(translate('Masstagging',
                                 '<b>Please choose a tagging profile.</b>'))
            self._startButton.setText(translate('Masstagging', '&Search'))
            return None

        tag_groups = split_files(self._status['selectedfiles'],
                                 mtp.file_pattern)

        search_msg = translate('Masstagging',
                               "An error occured during the search: <b>{}</b>")

        retrieve_msg = translate('Masstagging',
                                 "An error occured during album retrieval: <b>{}</b>")

        def search_error(error, mtp):
            thread.statusChanged.emit(search_msg.format(str(error)))

        def retrieval_error(error, mtp):
            thread.statusChanged.emit(retrieve_msg.format(str(error)))

        def run_masstag():
            replace_fields = []
            for files in tag_groups:
                mtp.clear()

                masstag(mtp, files, self.__flag, search_error,
                        retrieval_error)

                retrieved = merge_tsp_tracks(mtp.profiles)
                ret = match_files(files, retrieved,
                                  mtp.track_bound, mtp.fields,
                                  mtp.jfdi, mtp.leave_existing, True)[0]

                if ret:
                    thread.enable_preview_mode.emit()
                    thread.setpreview.emit(ret)

                set_status('<hr width="45%" /><br />')

        def finished(value):
            if not (value is True):
                set_status(translate('Masstagging',
                                     '<b>Lookups completed.</b>'))
            self._startButton.setText(translate('Masstagging', '&Search'))
            self.__flag.stop = False

        thread = PuddleThread(run_masstag, self)
        thread.setpreview.connect(self.setpreview)
        thread.enable_preview_mode.connect(self.enable_preview_mode)
        thread.threadfinished.connect(finished)
        thread.statusChanged.connect(set_status)

        thread.start()

    def writePreview(self):
        self.writepreview.emit()


control = ('Mass Tagging', MassTagWindow, RIGHTDOCK, False)

if __name__ == '__main__':
    app = QApplication([])
    from .. import puddletag, tagsources

    puddletag.load_plugins()
    sources = [z() for z in tagsources.tagsources]
    tsp = TagSourceProfile(None, sources[-1], ['field1', 'field2'],
                           1, ['repl1', 'repl2'])
    # win = TSProfileEdit(sources, tsp)
    mtp = MassTagProfile('Searching', 'Testing Search',
                         ['artist', 'title'], None, '%artist% - ktg',
                         [tsp], 0.70, 0.90, False, True,
                         {'album': [r'(.*?)\s+\(.*\)', '$1']})
    # win = MTProfileEdit(sources, mtp)

    win = MassTagEdit(sources)
    win.loadProfiles()
    win.show()
    app.exec_()



import csv
import os
import sys
from os.path import abspath, dirname, normcase, normpath, splitdrive
from os.path import join as path_join

from PyQt5.QtWidgets import QFileDialog, QApplication

from . import audioinfo
from .audioinfo.util import lnglength, encode_fn
from .findfunc import tagtofilename


def commonpath(a, b):
    """Returns the longest common to 'paths' path.

    Unlike the strange commonprefix:
    - this returns valid path
    - accepts only two arguments
    """
    a = normpath(normcase(a))
    b = normpath(normcase(b))

    if a == b:
        return a

    while len(a) > 0:
        if a == b:
            return a

        if len(a) > len(b):
            a = dirname(a)
        else:
            b = dirname(b)

    return None


def relpath(target, base_path=os.curdir):
    """\
    Return a relative path to the target from either the current directory
    or an optional base directory.

    Base can be a directory specified either as absolute or relative
    to current directory."""
    # http://code.activestate.com/recipes/302594/

    base_path = normcase(abspath(normpath(base_path)))
    target = normcase(abspath(normpath(target)))

    if base_path == target:
        return '.'

    # On the windows platform the target may be on a different drive.
    if splitdrive(base_path)[0] != splitdrive(target)[0]:
        return None

    common_path_len = len(commonpath(base_path, target))

    # If there's no common prefix decrease common_path_len should be less by 1
    base_drv, base_dir = splitdrive(base_path)
    if common_path_len == len(base_drv) + 1:
        common_path_len -= 1

    # if base_path is root directory - no directories up
    if base_dir == os.sep:
        dirs_up = 0
    else:
        dirs_up = base_path[common_path_len:].count(os.sep)

    ret = os.sep.join([os.pardir] * dirs_up)
    if len(target) > common_path_len:
        ret = path_join(ret, target[common_path_len + 1:])

    return ret


def readm3u(path):
    # From http://forums.fedoraforum.org/showthread.php?p=1224109
    fileHandle = open(path, 'r')
    reader = csv.reader(open(path, "r"))
    olddir = os.path.abspath(os.curdir)
    os.chdir(os.path.dirname(path))

    # List of mp3files
    mp3Files = []

    for row in reader:
        if len(row) < 1:
            # Skip blanks
            continue
        elif row[0].startswith("#"):
            # Ignore comments
            continue
        else:
            # store rule
            mp3Files.append(normpath(abspath(','.join(row))))

    fileHandle.close()
    os.chdir(olddir)
    return mp3Files


def exportm3u(tags, tofile, format=None, reldir=False, winsep=False):
    header = ['#EXTM3U']

    if reldir:
        reldir = os.path.dirname(os.path.abspath(tofile))
        filenames = [relpath(f.filepath, reldir) for f in tags]
    else:
        filenames = [f.filepath for f in tags]

    if winsep:
        filenames = [f.replace('/', '\\') for f in filenames]

    if format is None:
        text = '\n'.join(header + filenames)
    else:
        text = header
        extinfo = ('#EXTINF: %s, %s' % (str(lnglength(f.length)),
                                        encode_fn(tagtofilename(format, f, False))) for f in tags)
        [text.extend([z, y]) for z, y in zip(extinfo, filenames)]
        text = '\n'.join(text)

    playlist = open(tofile, 'w')
    playlist.write(text)
    playlist.close()


if __name__ == '__main__':
    app = QApplication(sys.argv)
    filedlg = QFileDialog()
    filedlg.setFileMode(QFileDialog.FileMode.Directory)
    filedlg.setOption(QFileDialog.Option.ShowDirsOnly)
    filename = str(filedlg.getExistingDirectory(None,
                                                'Open Folder'))
    tags = []
    for z in os.listdir(filename):
        try:
            tag = audioinfo.Tag(os.path.join(filename, z))
            if tag:
                tags.append(tag)
        except Exception as e:
            str(e)
    folder = str(filedlg.getSaveFileName(None,
                                         'Save File'))
    exportm3u(tags, folder)



# -*- coding: utf-8 -*-
import os
import sys
from copy import copy, deepcopy
from functools import partial

from PyQt5.QtCore import Qt, pyqtSignal
from PyQt5.QtWidgets import QAbstractItemView, QAction, QApplication, QCheckBox, QComboBox, QCompleter, \
    QDialog, QFrame, QGridLayout, QInputDialog, QLabel, QLineEdit, QListWidgetItem, QMenu, QMessageBox, \
    QScrollArea, QSizePolicy, QSpinBox, QStackedWidget, QToolButton, QVBoxLayout, QWidget
from pyparsing import delimited_list, alphanums, Combine, Word, QuotedString

from . import findfunc, functions
from . import functions_dialogs
from .audioinfo import INFOTAGS, READONLY
from .constants import (TEXT, COMBO, CHECKBOX, SAVEDIR, CONFIGDIR, ACTIONDIR)
from .findfunc import Function, apply_macros, apply_actions, Macro
from .puddleobjects import (ListBox, OKCancel, ListButtons, winsettings, gettaglist, settaglist, safe_name, open_resourcefile)
from .puddleobjects import PuddleConfig, PuddleCombo
from .puddleobjects import ShortcutEditor
from .util import (PluginFunction, translate, pprint_tag)

READONLY = list(READONLY)
FUNC_SETTINGS = os.path.join(CONFIGDIR, 'function_settings')

FIELDS_TOOLTIP = translate('Functions Dialog',
                           """<p>Fields that will
                           get written to.</p>
                       
                           <ul>
                           <li>Enter a list of comma-separated fields
                           eg. <b>artist, title, album</b></li>
                           <li>Use <b>__selected</b> to write only to the selected cells.
                           It is not allowed when creating an action.</li>
                           <li>Combinations like <b>__selected, artist, title</b> are
                           allowed.</li>
                           <li>But using <b>__selected</b> in Actions is <b>not</b>.</li>
                           <li>'~' will write to all the the fields, except what follows it
                           . Eg <b>~artist, title</b> will write to all but the artist and
                           title fields found in the selected files.<li>
                           </ul>""")


def displaytags(tags):
    text = pprint_tag(tags)
    if not text:
        return translate('Functions Dialog', '<b>No change.</b>')

    if text.endswith('<br />'):
        text = text[:-len('<br />')]
    return text


class ShortcutDialog(QDialog):
    shortcutChanged = pyqtSignal(str, name='shortcutChanged')

    def __init__(self, shortcuts=None, parent=None):
        super(ShortcutDialog, self).__init__(parent)
        self.setWindowTitle('puddletag')
        self.ok = False
        label = QLabel(translate('Shortcut Editor', 'Enter a key sequence for the shortcut.'))
        self._text = ShortcutEditor(shortcuts)

        okcancel = OKCancel()
        okcancel.cancelButton.setText(translate('Shortcut Editor', "&Don't assign keyboard shortcut."))
        okcancel.okButton.setEnabled(False)

        okcancel.ok.connect(self.okClicked)
        okcancel.cancel.connect(self.close)

        self._text.validityChanged.connect(okcancel.okButton.setEnabled)

        vbox = QVBoxLayout()
        vbox.addWidget(label)
        vbox.addWidget(self._text)
        vbox.addLayout(okcancel)
        vbox.addStretch()
        self.setLayout(vbox)

        self._shortcuts = shortcuts

    def okClicked(self):
        self.shortcutChanged.emit(str(self._text.text()))
        self.ok = True
        self.close()

    def getShortcut(self):
        self.exec_()
        if self._text.valid:
            return str(self._text.text()), self.ok
        else:
            return '', self.ok


class ShortcutName(QDialog):
    def __init__(self, texts, default='', parent=None):
        super(ShortcutName, self).__init__(parent)
        self.setWindowTitle('puddletag')
        self.ok = False
        self._texts = texts
        label = QLabel(translate('Actions', 'Enter a name for the shortcut.'))
        self._text = QLineEdit(default)

        okcancel = OKCancel()
        self._ok = okcancel.okButton
        self.enableOK(self._text.text())

        okcancel.ok.connect(self.okClicked)
        okcancel.cancel.connect(self.close)

        self._text.textChanged.connect(self.enableOK)

        vbox = QVBoxLayout()
        vbox.addWidget(label)
        vbox.addWidget(self._text)
        vbox.addLayout(okcancel)
        vbox.addStretch()
        self.setLayout(vbox)

    def okClicked(self):
        self.ok = True
        self.close()

    def enableOK(self, text):
        if text and str(text) not in self._texts:
            self._ok.setEnabled(True)
        else:
            self._ok.setEnabled(False)

    def getText(self):
        self.exec_()
        return str(self._text.text()), self.ok


class ScrollLabel(QScrollArea):
    def __init__(self, text='', parent=None):
        QScrollArea.__init__(self, parent)
        label = QLabel()
        label.setMargin(3)
        self.setWidget(label)
        self.setText(text)
        self.text = label.text
        label.setSizePolicy(QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Preferred)
        self.setFrameStyle(QFrame.Shape.NoFrame)
        self.setWidgetResizable(True)
        self.setAlignment(Qt.AlignmentFlag.AlignLeft | Qt.AlignmentFlag.AlignTop)
        self.setVerticalScrollBarPolicy(Qt.ScrollBarPolicy.ScrollBarAlwaysOff)

    def wheelEvent(self, e):
        h = self.horizontalScrollBar()
        if h.isVisible():
            numsteps = e.angleDelta().y() // 5
            h.setValue(h.value() - numsteps)
            e.accept()
        else:
            QScrollArea.wheelEvent(self, e)

    def setText(self, text):
        label = self.widget()
        label.setText(text)
        hbar = self.horizontalScrollBar()
        height = label.sizeHint().height() + hbar.height()
        self.setMaximumHeight(height)
        self.setMinimumHeight(height)


class FunctionDialog(QWidget):
    "A dialog that allows you to edit or create a Function class."

    _controls = {'text': PuddleCombo, 'combo': QComboBox, 'check': QCheckBox}

    signals = {
        TEXT: 'editTextChanged',
        COMBO: 'currentIndexChanged',
        CHECKBOX: 'stateChanged',
    }

    updateExample = pyqtSignal(object, name='updateExample')

    def __init__(self, funcname, selected_fields=False, userargs=None,
                 default_fields=None, parent=None, example=None, text=None):
        """funcname is name the function you want to use(can be either string, or functions.py function).
        if combotags is true then a combobox with tags that the user can choose from are shown.
        userargs is the default values you want to fill the controls in the dialog with
        [make sure they don't exceed the number of arguments of funcname]."""
        QWidget.__init__(self, parent)
        identifier = QuotedString('"') | Combine(Word
                                                 (alphanums + ' !"#$%&\'()*+-./:;<=>?@[\\]^_`{|}~'))
        tags = delimited_list(identifier)
        self.func = Function(funcname)
        docstr = self.func.doc[1:]
        self.vbox = QVBoxLayout()
        self.retval = []
        self._selectedFields = selected_fields

        if selected_fields:
            fields = ['__all'] + sorted(INFOTAGS) + \
                     selected_fields + gettaglist()
        else:
            fields = ['__selected', '__all'] + sorted(INFOTAGS) + \
                     gettaglist()

        self.tagcombo = QComboBox(self)
        self.tagcombo.setToolTip(FIELDS_TOOLTIP)
        self.tagcombo.setEditable(True)
        self.tagcombo.setCompleter(QCompleter(self.tagcombo))
        self.tagcombo.addItems(fields)

        self.tagcombo.editTextChanged.connect(self.showexample)

        if self.func.function not in functions.no_fields:
            label = QLabel(translate('Defaults', "&Fields"))
            self.vbox.addWidget(label)
            self.vbox.addWidget(self.tagcombo)
            label.setBuddy(self.tagcombo)
        else:
            self.tagcombo.setVisible(False)
        self.example = example
        self._text = text

        if self.func.function in functions_dialogs.dialogs:
            vbox = QVBoxLayout()
            vbox.addWidget(self.tagcombo)
            self.widget = functions_dialogs.dialogs[self.func.function](self)
            vbox.addWidget(self.widget)
            vbox.addStretch()
            self.setLayout(vbox)
            self.setMinimumSize(self.sizeHint())

            self.setArguments(default_fields, userargs)
            return
        else:
            self.widget = None

        self.textcombos = []
        # Loop that creates all the controls
        self.controls = []
        for argno, line in enumerate(docstr):
            args = tags.parse_string(line)
            label = args[0]
            ctype = args[1]
            default = args[2:]

            control, func, label = self._createControl(label, ctype, default)

            self.retval.append(func)
            self.controls.append(control)
            getattr(control, self.signals[ctype]).connect(self.showexample)

            if label:
                self.vbox.addWidget(label)
            self.vbox.addWidget(control)

        self.setArguments(default_fields, userargs)

        self.vbox.addStretch()
        self.setLayout(self.vbox)
        self.setMinimumSize(self.sizeHint())

    def argValues(self):
        """Returns the values in the windows controls.
        The last argument is the tags value.
        Also sets self.func's arg and tag values."""

        if self.widget:
            newargs = self.widget.arguments()
        else:
            newargs = []
            for method in self.retval:
                if method.__name__ == 'checkState':
                    if method() == Qt.CheckState.Checked:
                        newargs.append(True)
                    elif (method() == Qt.CheckState.PartiallyChecked) or (method() == Qt.CheckState.Unchecked):
                        newargs.append(False)
                else:
                    if isinstance(method(), int):
                        newargs.append(method())
                    else:
                        newargs.append(str(method()))
            [z.save() for z in self.textcombos]
        self.func.setArgs(newargs)

        fields = [z.strip() for z in
                  str(self.tagcombo.currentText()).split(",") if z]

        if self.func.function in functions.no_fields:
            self.func.setTag(['just nothing to do with this'])
        else:
            self.func.setTag(fields)
        return newargs + fields

    def _createControl(self, label, ctype, default=None):
        if ctype == 'text':
            control = self._controls['text'](label, parent=self)
        else:
            control = self._controls[ctype](self)

        if ctype == 'combo':
            func = control.currentText
            if default:
                control.addItems([translate('Functions', d) for d in default])
        elif ctype == 'text':
            self.textcombos.append(control)
            func = control.currentText
            if default:
                control.setEditText(default[0])
        elif ctype == 'check':
            func = control.checkState
            if default:
                if default[0] == "True" or default[0] is True:
                    control.setChecked(True)
                else:
                    control.setChecked(False)
            control.setText(translate('Functions', label))

        if ctype != 'check':
            label = QLabel(translate('Functions', label))
            label.setBuddy(control)
        else:
            label = None

        return control, func, label

    def loadSettings(self, filename=None):
        if filename is None:
            filename = FUNC_SETTINGS
        cparser = PuddleConfig(filename)
        function = self.func.function
        section = '%s_%s' % (function.__module__, function.__name__)
        arguments = cparser.get(section, 'arguments', [])
        fields = cparser.get(section, 'fields', [])
        if not fields:
            fields = None
        self.setArguments(fields, arguments)

    def saveSettings(self, filename=None):
        if not filename:
            filename = FUNC_SETTINGS
        function = self.func.function
        section = '%s_%s' % (function.__module__, function.__name__)

        cparser = PuddleConfig(filename)
        args = self.argValues()
        cparser.set(section, 'arguments', self.func.args)
        cparser.set(section, 'fields', self.func.tag)

    def showexample(self, *args, **kwargs):
        self.argValues()
        if self.example is not None:
            audio = self.example
            try:
                if self.func.function in functions.no_preview:
                    self.updateExample.emit(
                        translate('Functions Dialog',
                                  'No preview for is shown for this function.'))
                    return
                fields = findfunc.parse_field_list(self.func.tag, audio,
                                                   self._selectedFields)
                from .puddletag import status
                files = status['selectedfiles']
                files = str(len(files)) if files else '1'
                state = {'__counter': '0', '__total_files': files}
                val = apply_actions([self.func], audio, state, fields)
            except findfunc.ParseError as e:
                val = '<b>%s</b>' % (e.message)
            if val is not None:
                self.updateExample.emit(val)
            else:
                self.updateExample.emit(translate('Functions Dialog', '<b>No change</b>'))

    def _sanitize(self, ctype, value):
        if ctype in ['combo', 'text']:
            return value
        elif ctype == 'check':
            if value is True or value == 'True':
                return True
            else:
                return False
        elif ctype == 'spinbox':
            try:
                return int(value)
            except (TypeError, ValueError):
                return 0

    def setArguments(self, fields=None, args=None):
        if fields is not None:
            text = ', '.join(fields)
            index = self.tagcombo.findText(text)
            if index != -1:
                self.tagcombo.setCurrentIndex(index)
            else:
                self.tagcombo.insertItem(0, text)
                self.tagcombo.setCurrentIndex(0)
            self.tagcombo.setEditText(text)

        if not args:
            return

        if self.widget:
            self.widget.setArguments(*args)
            return

        for argument, control in zip(args, self.controls):
            if isinstance(control, QComboBox):
                index = control.findText(argument)
                if index != -1:
                    control.setCurrentIndex(index)
            elif isinstance(control, PuddleCombo):
                control.setEditText(argument)
            elif isinstance(control, QCheckBox):
                control.setChecked(self._sanitize('check', argument))
            elif isinstance(control, QSpinBox):
                control.setValue(self._sanitize('spinbox', argument))


class CreateFunction(QDialog):
    """A dialog to allow the creation of functions using only one window and a QStackedWidget.
    For each function in functions, a dialog is created and displayed in the stacked widget."""
    valschanged = pyqtSignal(object, name='valschanged')

    def __init__(self, prevfunc=None, selected_fields=None, parent=None,
                 example=None, text=None):
        """tags is a list of the tags you want to show in the FunctionDialog.
        Each item should be in the form (DisplayName, tagname) as used in audioinfo.
        prevfunc is a Function object that is to be edited."""
        QDialog.__init__(self, parent)
        self.setWindowTitle(translate('Functions Dialog', "Functions"))
        winsettings('createfunction', self)

        # Allow __selected field to be used.
        self.allowSelected = True

        self.realfuncs = []
        # Get all the function from the functions module.
        for z, funcname in functions.functions.items():
            if isinstance(funcname, PluginFunction):
                self.realfuncs.append(funcname)
            elif callable(funcname) and (not (funcname.__name__.startswith("__") or (funcname.__doc__ is None))):
                self.realfuncs.append(z)

        funcnames = [(Function(z).funcname, z) for z in self.realfuncs]
        funcnames.sort(key=lambda x: translate('Functions', x[0]))
        self.realfuncs = [z[1] for z in funcnames]

        self.vbox = QVBoxLayout()
        self.functions = QComboBox()
        self.functions.addItems(
            sorted([translate('Functions', x[0]) for x in funcnames]))
        self.vbox.addWidget(self.functions)

        self.stack = QStackedWidget()
        self.vbox.addWidget(self.stack)
        self.okcancel = OKCancel()

        self.stackWidgets = {}  # Holds the created windows in the form self.functions.index: window
        self.setLayout(self.vbox)
        self.setMinimumHeight(self.sizeHint().height())
        self.okcancel.ok.connect(self.okClicked)
        self.okcancel.cancel.connect(self.close)

        self.example = example
        self._text = text
        if not selected_fields:
            self.selectedFields = []
        else:
            self.selectedFields = selected_fields

        self.exlabel = ScrollLabel('')

        if prevfunc is not None:
            index = self.functions.findText(
                translate('Functions', prevfunc.funcname))
            if index >= 0:
                self.functions.setCurrentIndex(index)
                self.createWindow(index, prevfunc.tag, prevfunc.args)
        else:
            self.createWindow(0)

        self.functions.activated.connect(self.createWindow)

        self.vbox.addWidget(self.exlabel)
        self.vbox.addLayout(self.okcancel)
        self.setLayout(self.vbox)

    def createWindow(self, index, fields=None, args=None):
        """Creates a Function dialog in the stack window
        if it doesn't exist already."""
        self.stack.setFrameStyle(QFrame.Shape.Box)
        if index not in self.stackWidgets:
            widget = FunctionDialog(self.realfuncs[index],
                                    self.selectedFields, args, fields,
                                    example=self.example, text=self._text)
            if args is None:
                widget.loadSettings()
            self.stackWidgets.update({index: widget})
            self.stack.addWidget(widget)
            widget.updateExample.connect(self.updateExample)
        self.stack.setCurrentWidget(self.stackWidgets[index])
        self.stackWidgets[index].showexample()
        self.controls = getattr(self.stackWidgets[index], 'controls', [])
        self.setMinimumHeight(self.sizeHint().height())
        if self.sizeHint().width() > self.width():
            self.setMinimumWidth(self.sizeHint().width())

    def okClicked(self, close=True):
        w = self.stack.currentWidget()
        w.argValues()
        if not self.checkFields(w.func.tag):
            return

        if close:
            self.close()

        if w.func.tag:
            fields = gettaglist()
            new_fields = [z for z in w.func.tag if z not in fields]
            if new_fields:
                settaglist(sorted(new_fields + fields))

        for widget in self.stackWidgets.values():
            widget.saveSettings()
        self.saveSettings()
        self.valschanged.emit(w.func)

    def checkFields(self, fields):
        func = self.stack.currentWidget().func
        msg = translate('Actions',
                        "Error: Using <b>__selected</b> in Actions is not allowed.")
        if not self.allowSelected and '__selected' in fields:
            QMessageBox.warning(self, 'puddletag', msg)
            return False
        elif func is not None and func not in functions.no_fields:
            msg = translate('Actions',
                            "Please enter some fields to write to.")
            if not [_f for _f in fields if _f]:
                QMessageBox.information(self, 'puddletag', msg)
                return False
        return True

    def loadSettings(self):
        cparser = PuddleConfig()
        func_name = cparser.get('functions', 'last_used', '')
        if not func_name:
            return

        try:
            index = self.realfuncs.index(func_name)
            self.createWindow(index)
            self.functions.setCurrentIndex(index)
        except ValueError:
            return

    def saveSettings(self):
        cparser = PuddleConfig()
        funcname = self.realfuncs[self.functions.currentIndex()]
        cparser.set('functions', 'last_used', funcname)

    def updateExample(self, text):
        if not text:
            self.exlabel.setText('')
        else:
            self.exlabel.setText(displaytags(text))


class CreateAction(QDialog):
    "An action is defined as a collection of functions. This dialog serves the purpose of creating an action"
    donewithmyshit = pyqtSignal(list, name='donewithmyshit')

    def __init__(self, parent=None, prevfunctions=None, example=None):
        """tags is a list of the tags you want to show in the FunctionDialog.
        Each item should be in the form (DisplayName, tagname as used in audioinfo).
        prevfunction is the previous function that is to be edited."""
        QDialog.__init__(self, parent)
        self.setWindowTitle(translate('Actions', "Modify Action"))
        winsettings('editaction', self)
        self.grid = QGridLayout()

        self.listbox = ListBox()
        self.functions = []
        self.buttonlist = ListButtons()
        self.grid.addWidget(self.listbox, 0, 0)
        self.grid.addLayout(self.buttonlist, 0, 1)

        self.okcancel = OKCancel()
        self.setLayout(self.grid)
        self.example = example

        self.okcancel.cancel.connect(self.cancelClicked)
        self.okcancel.ok.connect(self.okClicked)
        self.buttonlist.add.connect(self.add)
        self.buttonlist.edit.connect(self.edit)
        self.buttonlist.moveup.connect(self.moveUp)
        self.buttonlist.movedown.connect(self.moveDown)
        self.buttonlist.remove.connect(self.remove)
        self.buttonlist.duplicate.connect(self.duplicate)
        self.listbox.currentRowChanged.connect(self.enableEditButtons)
        self.listbox.itemDoubleClicked.connect(self.edit)
        
        if len(self.functions) == 0:
            self.buttonlist.duplicateButton.setEnabled(False)
            self.buttonlist.editButton.setEnabled(False)

        if prevfunctions is not None:
            self.functions = copy(prevfunctions)
            self.listbox.addItems([function.description() for
                                   function in self.functions])

        if example:
            self._examplelabel = ScrollLabel('')
            self.grid.addWidget(self._examplelabel, 1, 0)
            self.grid.setRowStretch(0, 1)
            self.grid.setRowStretch(1, 0)
            self.example = example
            self.updateExample()
            self.grid.addLayout(self.okcancel, 2, 0, 1, 2)
        else:
            self.grid.addLayout(self.okcancel, 1, 0, 1, 2)
        self.enableOK()

    def updateExample(self):
        try:
            from .puddletag import status
            files = status['selectedfiles']
            files = str(len(files)) if files else '1'
            state = {'__counter': '0', '__total_files': files}
            tags = apply_actions(self.functions, self.example, state)
            self._examplelabel.setText(displaytags(tags))
        except findfunc.ParseError as e:
            self._examplelabel.setText(e.message)

    def enableEditButtons(self, val):
        if val == -1:
            [button.setEnabled(False) for button in self.buttonlist.widgets[1:]]
        else:
            [button.setEnabled(True) for button in self.buttonlist.widgets[1:]]

    def enableOK(self):
        if self.listbox.count() > 0:
            self.okcancel.okButton.setEnabled(True)
        else:
            self.okcancel.okButton.setEnabled(False)

    def moveDown(self):
        self.listbox.moveDown(self.functions)

    def moveUp(self):
        self.listbox.moveUp(self.functions)

    def remove(self):
        self.listbox.removeSelected(self.functions)
        self.updateExample()
        self.enableOK()

    def add(self):
        self.win = CreateFunction(None, parent=self, example=self.example)
        self.win.allowSelected = False
        self.win.setModal(True)
        self.win.show()
        self.win.valschanged.connect(self.addBuddy)

    def edit(self):
        self.win = CreateFunction(self.functions[self.listbox.currentRow()],
                                  parent=self, example=self.example)
        self.win.allowSelected = False
        self.win.setModal(True)
        self.win.show()
        self.win.valschanged.connect(self.editBuddy)

    def editBuddy(self, func):
        self.listbox.currentItem().setText(func.description())
        self.functions[self.listbox.currentRow()] = func
        self.updateExample()

    def addBuddy(self, func):
        self.listbox.addItem(func.description())
        self.functions.append(func)
        self.updateExample()
        self.enableOK()

    def okClicked(self):
        self.accept()
        self.close()
        self.donewithmyshit.emit(self.functions)

    def duplicate(self):
        self.win = CreateFunction(self.functions[self.listbox.currentRow()],
                                  parent=self, example=self.example)
        self.win.allowSelected = False
        self.win.setModal(True)
        self.win.show()
        self.win.valschanged.connect(self.addBuddy)

    def cancelClicked(self):
        self.reject()
        self.close()


class ActionWindow(QDialog):
    """Just a dialog that allows you to add, remove and edit actions
    On clicking OK, a signal "donewithmyshit" is emitted.
    It returns a list of lists.
    Each element of a list contains one complete action. While
    the elements of that action are just normal Function objects."""
    donewithmyshit = pyqtSignal(list, name='donewithmyshit')
    actionOrderChanged = pyqtSignal(name='actionOrderChanged')
    checkedChanged = pyqtSignal(list, name='checkedChanged')

    def __init__(self, parent=None, example=None, quickaction=None):
        """tags are the tags to be shown in the FunctionDialog"""
        QDialog.__init__(self, parent)
        self.setWindowTitle(translate('Actions', "Actions"))
        winsettings('actions', self)
        self._shortcuts = []
        self._quickaction = quickaction
        self.listbox = ListBox()
        self.listbox.setSelectionMode(QAbstractItemView.SelectionMode.ExtendedSelection)
        self.listbox.setEditTriggers(QAbstractItemView.EditTrigger.EditKeyPressed)

        self.example = example

        self.macros = self.loadMacros()
        cparser = PuddleConfig()
        self.__configKey = 'quick_actions' if quickaction else 'actions'
        to_check = cparser.get(self.__configKey, 'checked', [])

        for i, m in sorted(self.macros.items()):
            item = QListWidgetItem(m.name)
            item.setFlags(item.flags() | Qt.ItemFlag.ItemIsEditable)
            if m.name in to_check:
                item.setCheckState(Qt.CheckState.Checked)
            else:
                item.setCheckState(Qt.CheckState.Unchecked)
            self.listbox.addItem(item)

        self.okcancel = OKCancel()
        self.okcancel.okButton.setDefault(True)
        x = QAction(translate('Actions', 'Assign &Shortcut'), self)
        self.shortcutButton = QToolButton()
        self.shortcutButton.setDefaultAction(x)
        x.setToolTip(translate('Actions', '''<p>Creates a
            shortcut for the checked actions on the Actions menu.
            Use Edit Shortcuts (found by pressing down on this button)
            to edit shortcuts after the fact.</p>'''))
        menu = QMenu(self)
        edit_shortcuts = QAction(translate('Actions', 'Edit Shortcuts'), menu)
        edit_shortcuts.triggered.connect(self.editShortcuts)
        menu.addAction(edit_shortcuts)
        self.shortcutButton.setMenu(menu)

        self.okcancel.insertWidget(0, self.shortcutButton)
        self.grid = QGridLayout()

        self.buttonlist = ListButtons()

        self.grid.addWidget(self.listbox, 0, 0)
        self.grid.setRowStretch(0, 1)
        self.grid.addLayout(self.buttonlist, 0, 1)
        self.setLayout(self.grid)

        self.okcancel.ok.connect(self.okClicked)
        self.okcancel.cancel.connect(self.close)
        self.buttonlist.add.connect(self.add)
        self.buttonlist.edit.connect(self.edit)
        self.buttonlist.moveup.connect(self.moveUp)
        self.buttonlist.movedown.connect(self.moveDown)
        self.buttonlist.remove.connect(self.remove)
        self.buttonlist.duplicate.connect(self.duplicate)
        self.listbox.itemDoubleClicked.connect(self.edit)
        self.listbox.currentRowChanged.connect(self.enableListButtons)
        self.listbox.itemChanged.connect(self.renameAction)
        self.listbox.itemChanged.connect(self.enableOK)
        self.shortcutButton.clicked.connect(self.createShortcut)

        self._examplelabel = ScrollLabel('')
        self.grid.addWidget(self._examplelabel, 1, 0, 1, -1)
        self.grid.setRowStretch(1, 0)
        if example is None:
            self._examplelabel.hide()
        self.listbox.itemChanged.connect(self.updateExample)
        self.grid.addLayout(self.okcancel, 2, 0, 1, 2)
        self.updateExample()
        self.enableOK(None)

    def createShortcut(self):
        macros = self.checked()
        names = [m.name for m in macros]
        (name, ok) = ShortcutName(self.shortcutNames(), names[0]).getText()

        if name and ok:
            from . import puddletag
            shortcuts = [str(z.shortcut().toString()) for z in
                         puddletag.status['actions']]
            (shortcut, ok) = ShortcutDialog(shortcuts).getShortcut()
            name = str(name)

            from .action_shortcuts import (
                create_action_shortcut, save_shortcut)

            filenames = [m.filename for m in macros]

            if shortcut and ok:
                create_action_shortcut(name, filenames, shortcut, add=True)
            else:
                create_action_shortcut(name, filenames, add=True)
            save_shortcut(name, filenames)

    def editShortcuts(self):
        from . import action_shortcuts
        win = action_shortcuts.ShortcutEditor(True, self, True)
        win.setModal(True)
        win.show()

    def moveUp(self):
        self.listbox.moveUp(self.macros)

    def moveDown(self):
        self.listbox.moveDown(self.macros)

    def remove(self):
        cparser = PuddleConfig()
        listbox = self.listbox
        rows = sorted([listbox.row(item) for item in
                       listbox.selectedItems()])

        for row in rows:
            filename = self.macros[row].filename
            os.rename(filename, filename + '.deleted')
        self.listbox.removeSelected(self.macros)

        macros = {}
        for i, key in enumerate(self.macros):
            macros[i] = self.macros[key]

        macros = self.macros

        self.macros = dict((i, macros[k]) for i, k in
                           enumerate(sorted(macros)))

    def enableListButtons(self, val):
        if val == -1:
            [button.setEnabled(False) for button in self.buttonlist.widgets[1:]]
        else:
            [button.setEnabled(True) for button in self.buttonlist.widgets[1:]]

    def enableOK(self, val):
        item = self.listbox.item
        enable = [row for row in range(self.listbox.count()) if
                  item(row).checkState() == Qt.CheckState.Checked]
        if enable:
            self.okcancel.okButton.setEnabled(True)
            self.shortcutButton.setEnabled(True)
        else:
            self.okcancel.okButton.setEnabled(False)
            self.shortcutButton.setEnabled(False)

    def renameAction(self, item):
        name = str(item.text())
        names = [m.name for m in self.macros.values()]
        row = self.listbox.row(item)

        if name not in names:
            macro = self.macros[row]
            macro.name = name
            self.saveMacro(macro)
        else:
            self.listbox.blockSignals(True)
            item.setText(self.macros[row].name)
            self.listbox.blockSignals(False)

    def loadMacros(self):
        from glob import glob
        basename = os.path.basename

        funcs = {}
        cparser = PuddleConfig()
        set_value = partial(cparser.set, 'puddleactions')
        get_value = partial(cparser.get, 'puddleactions')

        firstrun = get_value('firstrun', True)
        set_value('firstrun', False)
        convert = get_value('convert', True)
        order = get_value('order', [])

        if convert:
            set_value('convert', False)
            findfunc.convert_actions(SAVEDIR, ACTIONDIR)
            if order:
                old_order = dict([(basename(z), i) for i, z in
                                  enumerate(order)])
                files = glob(os.path.join(ACTIONDIR, '*.action'))
                order = {}
                for i, action_fn in enumerate(files):
                    try:
                        order[old_order[basename(action_fn)]] = action_fn
                    except KeyError:
                        if not old_order:
                            order[i] = action_fn
                order = [z[1] for z in sorted(order.items())]
                set_value('order', order)

        files = glob(os.path.join(ACTIONDIR, '*.action'))
        if firstrun and not files:
            filenames = ['data:./caseconversion.action', 'data:./standard.action']
            files = list(map(open_resourcefile, filenames))
            set_value('firstrun', False)

            for fileobj, filename in zip(files, filenames):
                filename = os.path.join(ACTIONDIR, filename.rsplit('/', maxsplit=1)[1])
                f = open(filename, 'w')
                f.write(fileobj.read())
                f.close()
            files = glob(os.path.join(ACTIONDIR, '*.action'))

        files = [z for z in order if z in files] + \
                [z for z in files if z not in order]

        return dict((i, Macro(f)) for i, f in enumerate(files))

    def updateExample(self, *args):
        if self.example is None:
            self._examplelabel.hide()
            return
        l = self.listbox
        items = [l.item(z) for z in range(l.count())]
        selectedrows = [i for i, z in enumerate(items) if z.checkState() == Qt.CheckState.Checked]

        if selectedrows:
            from .puddletag import status
            files = status['selectedfiles']
            total = str(len(files)) if files else '1'
            state = {'__counter': '0', '__total_files': total}

            macros = [self.macros[i] for i in selectedrows]
            try:
                tags = apply_macros(macros, self.example, state,
                                    self._quickaction)
                self._examplelabel.setText(displaytags(tags))
            except findfunc.ParseError as e:
                self._examplelabel.setText(e.message)
            self._examplelabel.show()
        else:
            self._examplelabel.hide()

    def saveMacro(self, macro, filename=None):
        cparser = PuddleConfig()
        if filename is None and macro.filename:
            macro.save()
        elif filename:
            macro.filename = filename
            macro.save()
        else:
            name = macro.name
            filename = os.path.join(ACTIONDIR, safe_name(name) + '.action')
            base = os.path.splitext(filename)[0]
            i = 0
            while os.path.exists(filename):
                filename = "%s_%d" % (base, i) + '.action'
                i += 1
            macro.save(filename)
            macro.filename = filename
        return filename

    def add(self):
        (text, ok) = QInputDialog.getText(self,
                                          translate('Actions', "New Action"),
                                          translate('Actions', "Enter a name for the new action."),
                                          QLineEdit.EchoMode.Normal)

        if (ok is True) and text:
            item = QListWidgetItem(text)
            item.setCheckState(Qt.CheckState.Unchecked)
            item.setFlags(item.flags() | Qt.ItemFlag.ItemIsEditable)
            self.listbox.addItem(item)
        else:
            return
        win = CreateAction(self, example=self.example)
        win.setWindowTitle(translate('Actions', "Add Action: ") + \
                           self.listbox.item(self.listbox.count() - 1).text())
        win.setModal(True)
        win.donewithmyshit.connect(self.addBuddy)
        win.rejected.connect(lambda: self.listbox.takeItem(self.listbox.count() - 1))
        win.show()

    def addBuddy(self, actions):
        m = Macro()
        m.name = str(self.listbox.item(self.listbox.count() - 1).text())
        m.actions = actions
        self.saveMacro(m)
        self.macros[self.listbox.count() - 1] = m

    def edit(self):
        m = self.macros[self.listbox.currentRow()]
        win = CreateAction(self, m.actions, example=self.example)
        win.setWindowTitle(
            translate('Actions', "Edit Action: ") + m.name)
        win.show()
        win.donewithmyshit.connect(self.editBuddy)

    def editBuddy(self, actions):
        m = self.macros[self.listbox.currentRow()]
        m.actions = actions
        self.saveMacro(m)
        self.updateExample()

    def checked(self):
        return [self.macros[row] for row in self.checkedRows()]

    def checkedRows(self):
        l = self.listbox
        items = [l.item(z) for z in range(l.count())]
        checked = [i for i, z in enumerate(items) if
                   z.checkState() == Qt.CheckState.Checked]
        return checked

    def saveChecked(self):
        cparser = PuddleConfig()
        m_names = [m.name for m in self.checked()]
        cparser.set(self.__configKey, 'checked', m_names)

    def saveOrder(self):
        macros = self.macros
        cparser = PuddleConfig()
        order = [macros[i].filename for i in sorted(macros)]
        lastorder = cparser.get('puddleactions', 'order', [])
        if lastorder == order:
            return
        cparser.set('puddleactions', 'order', order)
        self.actionOrderChanged.emit()

    def close(self):
        self.saveOrder()
        QDialog.close(self)

    def okClicked(self, close=True):
        """When clicked, save the current contents of the listbox and the associated functions"""
        macros = self.checked()
        names = [m.name for m in macros]
        cparser = PuddleConfig()
        cparser.set(self.__configKey, 'checked', names)
        if close:
            self.close()

        self.checkedChanged.emit(self.checkedRows())
        self.donewithmyshit.emit(macros)

    def duplicate(self):
        l = self.listbox
        if len(l.selectedItems()) > 1:
            return
        row = l.currentRow()
        oldname = self.macros[row].name

        (text, ok) = QInputDialog.getText(self,
                                          translate('Actions', "Copy %s action" % oldname),
                                          translate('Actions', "Enter a name for the new action."),
                                          QLineEdit.EchoMode.Normal)
        if not (ok and text):
            return

        name = str(text)
        actions = deepcopy(self.macros[row].actions)

        win = CreateAction(self, actions, example=self.example)
        win.setWindowTitle(
            translate('Actions', "Edit Action: {}").format(name))

        win.show()
        dupebuddy = partial(self.duplicateBuddy, name)
        win.donewithmyshit.connect(dupebuddy)

    def duplicateBuddy(self, name, actions):
        item = QListWidgetItem(name)
        item.setCheckState(Qt.CheckState.Unchecked)
        item.setFlags(item.flags() | Qt.ItemFlag.ItemIsEditable)
        self.listbox.addItem(item)

        m = Macro()
        m.name = name
        m.actions = actions
        self.saveMacro(m)
        self.macros[self.listbox.count() - 1] = m

    def shortcutNames(self):
        from .action_shortcuts import load_settings
        return [name for name, filename in load_settings()[1]]


if __name__ == "__main__":
    app = QApplication(sys.argv)
    app.setOrganizationName("Puddle Inc.")
    app.setApplicationName("puddletag")
    qb = ActionWindow([('Path', '__path'), ('Artist', 'artist'), ('Title', 'title'), ('Album', 'album'), ('Track', 'track'), ('Length', '__length'), ('Year', 'date')])
    qb.show()
    app.exec_()



# -*- coding: utf-8 -*-

from setuptools import setup

def _runtime_dependencies():
    """ Read the runtime dependencies from the requirements file """
    with open('requirements.txt') as f:
        return f.read().splitlines()

def _version():
    """ Read the version from the puddlestuff package """
    version = {}
    with open("puddlestuff/__init__.py") as fp:
        exec(fp.read(), version)
    return version['version_string']


def readme():
    with open('README.md') as f:
        return f.read()

setup(
    name='puddletag',
    version=_version(),
    author='puddletag developers',
    url='https://docs.puddletag.net/',
    download_url='https://github.com/puddletag/puddletag',
    description='Powerful, simple, audio tag editor',
    long_description=readme(),
    long_description_content_type='text/markdown',
    packages=['puddlestuff',
              'puddlestuff.mainwin',
              'puddlestuff.libraries',
              'puddlestuff.audioinfo',
              'puddlestuff.tagsources',
              'puddlestuff.tagsources.mp3tag',
              'puddlestuff.masstag',
              'puddlestuff.plugins'],
    package_data={'puddlestuff': ['data/*', 'translations/*.qm']},
    keywords='tagging ogg mp3 apev2 mp4 id3',
    license='GNU General Public License v3 or later (GPLv3+)',
    classifiers=['Development Status :: 5 - Production/Stable',
                 'Intended Audience :: End Users/Desktop',
                 'Natural Language :: English',
                 'Operating System :: POSIX :: Linux',
                 'Programming Language :: Python :: 3 :: Only',
                 'Topic :: Multimedia :: Sound/Audio :: Editors',
                 ],
    scripts=['puddletag'],
    python_requires=">=3.7",
    install_requires=_runtime_dependencies(),
    data_files=[('share/pixmaps/', ['puddletag.png', ]),
                ('share/applications/', ['puddletag.desktop', ]),
                ('share/man/man1/', ['puddletag.1', ])]
)



import tags
from PyQt5.QtCore import pyqtSignal
from PyQt5.QtWidgets import QInputDialog, QPushButton, QVBoxLayout, QWidget

from .. import audioinfo
from ..audioinfo.util import (PATH,
                              FILENAME, INFOTAGS,
                              READONLY, isempty, FILETAGS, EXTENSION, DIRPATH,
                              getdeco, setdeco)
from ..constants import RIGHTDOCK
from ..puddleobjects import PuddleDock

ATTRIBUTES = ('frequency', 'length', 'bitrate', 'accessed', 'size', 'created',
              'modified')


class Tag(audioinfo.MockTag):
    """Use as base for all tag classes."""
    IMAGETAGS = ()
    mapping = {}
    revmapping = {}

    _hash = {PATH: 'filepath',
             FILENAME: 'filename',
             EXTENSION: 'ext',
             DIRPATH: 'dirpath'}

    def __init__(self, dictionary=None):
        # self.IMAGETAGS = tuple(set([IMAGETAGS[random.randint(0,3)] for z in IMAGETAGS]))
        self.IMAGETAGS = ()
        self.images = []
        # if audioinfo.DATA not in self.IMAGETAGS:
        # self.IMAGETAGS = (,)
        # else:
        # images = set([random.randint(0, 11) for z in xrange(11)])
        # self.images = [self.image(**{'data': pictures[i],
        # 'imagetype': random.randint(0,21),
        # 'description': unicode(i)}) for i in images]

        if dictionary:
            self.link(dictionary)

        self._set_attrs(ATTRIBUTES)

    def link(self, dictionary):
        self._tags = dictionary
        self.filepath = dictionary['__filename']

    def save(self):
        logging.info('saving ' + self.filename)

    def copy(self):
        return Tag(self._tags.copy())

    @getdeco
    def __getitem__(self, key):
        """Get the tag value from self._tags. There is a slight
        caveat in that this method will never return a KeyError exception.
        Rather it'll return ''."""
        if key == '__image':
            return self.images

        try:
            return self._tags[key]
        except KeyError:
            # This is a bit of a bother since there will never be a KeyError exception
            # But its needed for the sort method in tagmodel.TagModel, .i.e it fails
            # if a key doesn't exist.
            return ""

    @setdeco
    def __setitem__(self, key, value):
        if key in READONLY:
            return
        elif key in FILETAGS:
            setattr(self, self._hash[key], value)
            return

        if key not in INFOTAGS and isempty(value):
            del (self[key])
        elif key in INFOTAGS or isinstance(key, int):
            self._tags[key] = value
        elif (key not in INFOTAGS) and isinstance(value, (str, int)):
            self._tags[key.lower()] = [str(value)]
        else:
            self._tags[key.lower()] = [str(z) for z in value]

    def mutvalues(self):
        # Retrieves key, value pairs according to id3.
        return [self._tags[key] for key in self if type(key) is not int and not key.startswith('__')]


class TestWidget(QWidget):
    setpreview = pyqtSignal(dict, name='setpreview')

    def __init__(self, parent=None, status=None):
        QWidget.__init__(self, parent)
        self.emits = ['setpreview']
        self.receives = []
        button = QPushButton('Set Previews')
        button.clicked.connect(self._changePreview)
        self._status = status

        load1000button = QPushButton('Load 1000')
        load1000button.clicked.connect(self._load1000)
        self._status = status

        save_tags = QPushButton('Save Files')
        save_tags.clicked.connect(self._saveTags)
        self._status = status

        load_many = QPushButton('Load Many')
        load_many.clicked.connect(self._loadMany)
        self._status = status

        box = QVBoxLayout()
        box.addWidget(button)
        box.addWidget(load1000button)
        box.addWidget(save_tags)
        box.addWidget(load_many)
        self.setLayout(box)

    def _changePreview(self):
        status = self._status
        files = status['allfiles'][2:10]
        d = {}
        for f in files:
            d[f] = {'artist': 'Preview Artist', 'title': 'Preview Title'}
        self.setpreview.emit(d)

    def _load1000(self):
        table = PuddleDock._controls['table']
        table.model().load(list(map(Tag, tags.tags)))
        table.model().saveModification = False

    def _loadMany(self):
        model = PuddleDock._controls['table'].model()
        num, ok = QInputDialog.getInt(self, 'puddletag',
                                      'Enter the number of files to fill the file-view with.', 10)
        if num:
            # pdb.set_trace()
            model.load(list(map(Tag, tags.tags[:num])))
            model.saveModification = False

    def _saveTags(self):
        files = self._status['allfiles']
        f = open('savedfiles', 'w')
        f.write('# -*- coding: utf-8 -*-\ntags = [%s]' %
                ',\n'.join((str(z.tags) for z in files)))
        f.close()


control = ('Puddle Testing', TestWidget, RIGHTDOCK, False)



import os

from PyQt5.QtCore import QDir, QDirIterator, QItemSelectionModel, QMutex, QSettings, QUrl, Qt, pyqtSignal
from PyQt5.QtGui import QDesktopServices
from PyQt5.QtWidgets import QAbstractItemView, QAction, QCheckBox, QFileSystemModel, QHeaderView, QMenu, QTreeView, QVBoxLayout, QWidget

from ..constants import LEFTDOCK, QT_CONFIG
from ..puddleobjects import (PuddleConfig, PuddleThread,
                             PuddleHeader)
from ..puddlesettings import (load_gen_settings,
                              save_gen_settings)
from ..tagmodel import has_previews
from ..translations import translate

qmutex = QMutex()

class DirModel(QFileSystemModel):
    def __init__(self, parent=None):
        QFileSystemModel.__init__(self, parent)

        self.setRootPath('/')
        self.setFilter(QDir.Filter.Dirs | QDir.Filter.NoDotAndDotDot)
        self.setReadOnly(False)

    def hasChildren(self, parent):
        """Overridden to only show the expand arrow when there are subfolders."""
        if parent.flags() & Qt.ItemFlag.ItemNeverHasChildren:
            return False
        elif (self.filePath(parent)):
            return QDirIterator(
                self.filePath(parent),
                self.filter() | QDir.Filter.NoDotAndDotDot,
                QDirIterator.IteratorFlag.NoIteratorFlags
            ).hasNext()
        return QFileSystemModel.hasChildren(self, parent)

class DirView(QTreeView):
    """The treeview used to select a directory."""
    removeFolders = pyqtSignal(list, bool, name='removeFolders')
    loadFiles = pyqtSignal(object, list, bool, name='loadFiles')

    def __init__(self, parent=None, subfolders=False, status=None):
        QTreeView.__init__(self, parent)

        self._load = False  # If True a loadFiles signal is emitted when
        # an index is clicked. See selectionChanged.

        header = PuddleHeader(Qt.Orientation.Horizontal, self)
        self.setHeader(header)
        self.header().setSectionResizeMode(QHeaderView.ResizeMode.ResizeToContents)

        self.setModel(DirModel())
        [self.hideColumn(column) for column in range(1, 4)]

        self.header().hide()
        self.subfolders = subfolders
        self.setSelectionMode(QAbstractItemView.SelectionMode.ExtendedSelection)
        self._lastselection = 0  # If > 0 appends files. See selectionChanged
        self._load = True
        self.setDragEnabled(False)
        self.setAcceptDrops(True)
        self.setDropIndicatorShown(True)
        self._dropaction = Qt.DropAction.MoveAction
        self._threadRunning = False

        self._select = True

        self.expanded.connect(lambda discarded: self.resizeColumnToContents(0))

    def checkPreviews(self, deselected):
        """Confirm to user if any file have previewes.

        If any currently loaded files have previews then the user
        is asked to confirm whether they want to change to a new dir.
        Returns True if the user cancels the action, False if
        the user chooses to go ahead or there are no un-commited previews.
        """
        msg = translate('Previews', 'Some files have uncommited previews. '
                                    'Changes will be lost once you load a directory. <br />'
                                    'Do you still want to load a new directory?<br />')
        if not has_previews(parent=self.parentWidget(), msg=msg):
            return False
        select = self._select
        self._select = False
        smodel = self.selectionModel()
        smodel.blockSignals(True)
        smodel.clear()
        smodel.select(deselected, QItemSelectionModel.SelectionFlag.Select)
        smodel.blockSignals(False)
        self._select = select
        return True

    def clearSelection(self, *args):
        self.blockSignals(True)
        self.selectionModel().clearSelection()
        self.blockSignals(False)
        self.removeFolders.emit([], True)

    def contextMenuEvent(self, event):

        connect = lambda o, s: o.triggered.connect(s)

        menu = QMenu(self)
        refresh = QAction(translate("Dirview",
                                    'Refresh Directory'), self)

        index = self.indexAt(event.pos())
        connect(refresh, lambda: self.model().fetchMore(index))

        header = self.header()
        if self.header().isHidden():
            show_header = QAction(translate("Dirview",
                                            'Show Header'), self)
            connect(show_header, header.show)
        else:
            show_header = QAction(translate("Dirview",
                                            'Hide Header'), self)
            connect(show_header, header.hide)

        open_dir = QAction(translate(
            'Dirview', 'Open in File Manager'), self)
        connect(open_dir, lambda: self.openExtern(index))

        menu.addAction(refresh)
        menu.addAction(show_header)
        menu.addAction(open_dir)

        menu.exec_(event.globalPos())
        super(DirView, self).contextMenuEvent(event)

    def dirMoved(self, dirs):
        if not self.isVisible():
            return
        self._load = False
        model = self.model()
        selectindex = self.selectionModel().select
        getindex = model.index
        exists = os.path.exists

        parents = set([os.path.dirname(z[0]) for z in dirs])

        def get_str(f):
            return model.filePath(f)

        selected = list(map(get_str, self.selectedIndexes()))

        for p in parents:
            if exists(p):
                i = getindex(p)
                while not i.isValid():
                    p = os.path.dirname(p)
                    i = getindex(p)
                model.fetchMore(i)

        for d in [z[1] for z in dirs] + selected:
            self.selectIndex(getindex(d))
        self._load = True

    def loadSettings(self):
        settings = QSettings(QT_CONFIG, QSettings.Format.IniFormat)
        header = self.header()
        if settings.value('dirview/header'):
            header.restoreState(settings.value('dirview/header'))
        hide = bool(settings.value('dirview/hide', True))
        self.setHeaderHidden(hide)

        if self.isVisible() == False:
            return

        cparser = PuddleConfig()
        d = cparser.get('main', 'lastfolder', '/')
        while not os.path.exists(d):
            d = os.path.dirname(d)
            if not d:
                return

        def expand_thread_func():
            index = self.model().index(d)
            parents = []
            while index.isValid():
                parents.append(index)
                index = index.parent()
            return parents

        def expandindexes(indexes):
            self.setEnabled(False)
            [self.expand(index) for index in indexes]
            self.setEnabled(True)

        thread = PuddleThread(expand_thread_func, self)
        thread.threadfinished.connect(expandindexes)
        thread.start()

    def mousePressEvent(self, event):
        if event.buttons() == Qt.MouseButton.RightButton:
            return
        else:
            super(DirView, self).mousePressEvent(event)

    def openExtern(self, index):
        if index.isValid():
            filename = self.model().filePath(index)
            QDesktopServices.openUrl(QUrl.fromLocalFile(filename))

    def selectDirs(self, dirlist):
        if self._threadRunning:
            return
        if not self.isVisible():
            return
        if not self._select:
            self._select = True
            return
        load = self._load
        self._load = False
        if not dirlist:
            self._load = False
            self.selectionModel().clear()
            self._load = load
            return

        if isinstance(dirlist, str):
            dirlist = [dirlist]
        self._threadRunning = True
        self.setEnabled(False)
        self.selectionModel().clear()
        selectindex = self.selectionModel().select
        getindex = self.model().index
        parent = self.model().parent

        def func():
            toselect = []
            toexpand = []
            for d in dirlist:
                if not os.path.exists(d):
                    continue
                if isinstance(d, bytes):
                    try:
                        d = str(d, 'utf8')
                    except (UnicodeEncodeError, UnicodeDecodeError):
                        pass
                index = getindex(d)
                toselect.append(index)
                i = parent(index)
                parents = []
                while i.isValid():
                    parents.append(i)
                    i = parent(i)
                toexpand.extend(parents)
            return (toselect, toexpand)

        def finished(val):
            qmutex.lock()
            select = val[0]
            expand = val[1]
            if select:
                self.setCurrentIndex(select[0])
                self.scrollTo(select[0])
            [selectindex(z, QItemSelectionModel.SelectionFlag.Select) for z in select]
            if expand:
                [self.expand(z) for z in expand]
            self.blockSignals(False)
            self.setEnabled(True)
            self._load = load
            self._threadRunning = False
            qmutex.unlock()

        dirthread = PuddleThread(func, self)
        dirthread.threadfinished.connect(finished)
        dirthread.start()

    def saveSettings(self):
        settings = QSettings(QT_CONFIG, QSettings.Format.IniFormat)
        settings.setValue('dirview/header',
                          self.header().saveState())
        settings.setValue('dirview/hide', self.isHeaderHidden())

    def selectionChanged(self, selected, deselected):
        QTreeView.selectionChanged(self, selected, deselected)
        if not self._load:
            self._lastselection = len(self.selectedIndexes())
            return

        getfilename = self.model().filePath
        dirs = list(set([getfilename(i) for
                         i in selected.indexes()]))
        old = list(set([getfilename(i) for
                        i in deselected.indexes()]))
        if self._lastselection:
            if len(old) == self._lastselection:
                append = False
            else:
                append = True
        else:
            append = False
        dirs = list(set(dirs).difference(old))
        if old:
            self.removeFolders.emit(old, False)

        if self.checkPreviews(deselected):
            return
        if dirs:
            self.loadFiles.emit(None, dirs, append)
        self._lastselection = len(self.selectedIndexes())
        self._select = False

    def selectIndex(self, index):
        if not index.isValid():
            return
        self.selectionModel().select(index, QItemSelectionModel.SelectionFlag.Select)
        parent = index.parent()
        while parent.isValid():
            self.expand(index)
            parent = parent.parent()

    def focusDir(self, dir_path):
        """Focuses the component on the given path """
        self.selectDirs(dir_path)


class DirViewWidget(QWidget):
    loadFiles = pyqtSignal(object, list, bool, name='loadFiles')
    removeFolders = pyqtSignal(list, bool, name='removeFolders')

    def __init__(self, parent=None, subfolders=False, status=None):
        super(DirViewWidget, self).__init__(parent)
        self._status = status

        self.dirview = DirView(self, subfolders, status)
        self.dirview.loadFiles.connect(self.loadFiles)
        self.dirview.removeFolders.connect(self.removeFolders)

        self.receives = [
            ('dirschanged', self.dirview.selectDirs),
            ('dirsmoved', self.dirview.dirMoved),
        ]
        self.emits = ['loadFiles', 'removeFolders']

        self.subfolderCheck = QCheckBox(translate('Dirview', 'Subfolders'),
                                        self)
        self.subfolderCheck.stateChanged.connect(self.setSubFolders)

        layout = QVBoxLayout()
        layout.addWidget(self.dirview, 1)
        layout.addWidget(self.subfolderCheck, 0)
        self.setLayout(layout)

    def loadSettings(self):
        self.dirview.loadSettings()
        self.subfolderCheck.blockSignals(True)
        if load_gen_settings([('Su&bfolders', True)])[0][1]:
            self.subfolderCheck.setChecked(True)
        else:
            self.subfolderCheck.setChecked(False)
        self.subfolderCheck.blockSignals(False)

    def saveSettings(self):
        self.dirview.saveSettings()

    def setSubFolders(self, check):
        if check == Qt.CheckState.Checked:
            value = True
        else:
            value = False
        self.dirview.subfolders = value
        save_gen_settings({'Su&bfolders': value})
        self._status['table'].subFolders = value

    def _focusDir(self, dir_path):
        """Focuses the Filesystem component on the given path """
        self.parentWidget().show()
        self.dirview.focusDir(dir_path)


    def focusCurrentFileParent(parent=None):
        """Focuses the Filesystem component on the parent folder of the currently selected file """

        files = parent._status['selectedfiles']
        if not files:
            return

        dir_path = files[0].dirpath
        parent._focusDir(dir_path)


control = ('Filesystem', DirViewWidget, LEFTDOCK, True)



import glob
import os

from ..constants import CONFIGDIR
from ..masstag import (fields_from_text, MassTagProfile,
                       TagSourceProfile)
from ..puddleobjects import PuddleConfig
from ..audioinfo import encode_fn

PROFILEDIR = os.path.join(CONFIGDIR, 'masstagging')
CONFIG = os.path.join(CONFIGDIR, 'masstagging.conf')

ALBUM_BOUND = 'album_bound'
TRACK_BOUND = 'track_bound'
PATTERN = 'pattern'
FIELDS = 'fields'
JFDI = 'jfdi'
NAME = 'name'
DESC = 'desc'
EXISTING_ONLY = 'leave_existing'
REPLACE_FIELDS = 'replace_fields'


class DummyTS(object):
    group_by = []


def convert_mtps(dirpath=PROFILEDIR):
    for fn in glob.glob(PROFILEDIR + '/*.conf'):
        mtp = convert_mtp(fn)
        os.rename(fn, os.path.splitext(fn)[0] + '.old')
        save_mtp(mtp, os.path.join(dirpath, encode_fn(mtp.name)) + '.mtp')


def convert_mtp(filename):
    cparser = PuddleConfig(filename)
    info_section = 'info'
    name = cparser.get(info_section, NAME, '')
    numsources = cparser.get(info_section, 'numsources', 0)
    album_bound = cparser.get(info_section, ALBUM_BOUND, 70) / 100.0
    track_bound = cparser.get(info_section, TRACK_BOUND, 80) / 100.0
    match_fields = cparser.get(info_section, FIELDS, ['artist', 'title'])
    pattern = cparser.get(info_section, PATTERN,
                          '%artist% - %album%/%track% - %title%')
    jfdi = cparser.get(info_section, JFDI, True)
    desc = cparser.get(info_section, DESC, '')
    existing = cparser.get(info_section, EXISTING_ONLY, False)

    ts_profiles = []
    for num in range(numsources):
        section = 'config%s' % num
        get = lambda key, default: cparser.get(section, key, default)

        source = DummyTS()
        source.name = get('source', '')
        fields = fields_from_text(get('fields', ''))
        no_result = get('no_match', 0)

        ts_profiles.append(TagSourceProfile(None, source, fields,
                                            no_result))

    return MassTagProfile(name, desc, match_fields, None,
                          pattern, ts_profiles, album_bound, track_bound, jfdi, existing,
                          '')


def load_all_mtps(dirpath=PROFILEDIR, tag_sources=None):
    if tag_sources is None:
        tag_sources = {}
    mtps = [mtp_from_file(fn, tag_sources) for fn
            in glob.glob(dirpath + '/*.mtp')]
    try:
        order = open(os.path.join(dirpath, 'order'), 'r').read().split('\n')
    except EnvironmentError:
        return mtps

    order = [z.strip() for z in order]
    first = []
    last = []

    names = dict([(mtp.name, mtp) for mtp in mtps])
    mtps = [names[name] for name in order if name in names]
    mtps.extend([names[name] for name in names if name not in order])

    return mtps


def mtp_from_file(filename=CONFIG, tag_sources=None):
    if tag_sources is None:
        tag_sources = {}
    else:
        tag_sources = dict((z.name, z) for z in tag_sources)

    cparser = PuddleConfig(filename)
    info_section = 'info'

    name = cparser.get(info_section, NAME, '')
    numsources = cparser.get(info_section, 'numsources', 0)
    album_bound = cparser.get(info_section, ALBUM_BOUND, 70) / 100.0
    track_bound = cparser.get(info_section, TRACK_BOUND, 80) / 100.0
    match_fields = cparser.get(info_section, FIELDS, ['artist', 'title'])
    pattern = cparser.get(info_section, PATTERN,
                          '%artist% - %album%/%track% - %title%')
    jfdi = cparser.get(info_section, JFDI, True)
    desc = cparser.get(info_section, DESC, '')
    leave_existing = cparser.get(info_section, EXISTING_ONLY, False)
    regexps = ''

    ts_profiles = []
    for num in range(numsources):
        section = 'config%s' % num
        get = lambda key, default: cparser.get(section, key, default)

        source = tag_sources.get(get('source', ''), None)
        no_result = get('no_match', 0)
        fields = fields_from_text(get('fields', ''))
        replace_fields = fields_from_text(get('replace_fields', ''))

        ts_profiles.append(TagSourceProfile(None, source, fields,
                                            no_result, replace_fields))

    mtp = MassTagProfile(name, desc, match_fields, None,
                         pattern, ts_profiles, album_bound, track_bound, jfdi,
                         leave_existing, regexps)

    return mtp


def save_mtp(mtp, filename=CONFIG):
    cparser = PuddleConfig(filename)
    info_section = 'info'

    cparser.set(info_section, NAME, mtp.name)
    for key in ['name', 'desc', 'file_pattern', 'fields',
                'jfdi', 'leave_existing']:
        cparser.set(info_section, key, getattr(mtp, key))

    for key in ['album_bound', 'track_bound']:
        cparser.set(info_section, key, int(getattr(mtp, key) * 100))

    cparser.set(info_section, 'numsources', len(mtp.profiles))
    for num, tsp in enumerate(mtp.profiles):
        section = 'config%s' % num
        name = tsp.tag_source.name if tsp.tag_source else ''
        cparser.set(section, 'source', name)
        cparser.set(section, 'if_no_result', tsp.if_no_result)
        cparser.set(section, 'fields', ','.join(tsp.fields))
        cparser.set(section, 'replace_fields', ','.join(tsp.replace_fields))


if __name__ == '__main__':
    fns = glob.glob(PROFILEDIR + '/Local.conf')
    convert_mtps(PROFILEDIR)
